# Copyright 2021-2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for connecting to the IQM quantum computer server interface."""

from __future__ import annotations

from collections.abc import Callable, Iterable
from datetime import datetime
from importlib.metadata import version
import itertools
import json
import os
import platform
import time
from typing import Any, TypeVar
from uuid import UUID
import warnings

from iqm.iqm_client.api import APIConfig, APIEndpoint, APIVariant
from iqm.iqm_client.authentication import TokenManager
from iqm.iqm_client.errors import (
    APITimeoutError,
    CircuitExecutionError,
    CircuitValidationError,
    ClientAuthenticationError,
    ClientConfigurationError,
    EndpointRequestError,
    JobAbortionError,
)
from iqm.iqm_client.models import (
    _SUPPORTED_OPERATIONS,
    CalibrationSet,
    Circuit,
    CircuitBatch,
    CircuitCompilationOptions,
    ClientLibrary,
    ClientLibraryDict,
    DictDict,
    DynamicQuantumArchitecture,
    Instruction,
    MoveGateValidationMode,
    QualityMetricSet,
    QuantumArchitecture,
    QuantumArchitectureSpecification,
    RunCounts,
    RunRequest,
    RunResult,
    RunStatus,
    StaticQuantumArchitecture,
    Status,
    serialize_qubit_mapping,
    validate_circuit,
)
from packaging.version import parse
from pydantic import BaseModel
from pydantic import ValidationError as PydanticValidationError
import requests
from requests import HTTPError

T_BaseModel = TypeVar("T_BaseModel", bound=BaseModel)

REQUESTS_TIMEOUT = float(os.environ.get("IQM_CLIENT_REQUESTS_TIMEOUT", 120.0))
DEFAULT_TIMEOUT_SECONDS = 900
SECONDS_BETWEEN_CALLS = float(os.environ.get("IQM_CLIENT_SECONDS_BETWEEN_CALLS", 1.0))


class IQMClient:
    """Provides access to IQM quantum computers.

    Args:
        url: Endpoint for accessing the server. Has to start with http or https.
        client_signature: String that IQMClient adds to User-Agent header of requests
            it sends to the server. The signature is appended to IQMClient's own version
            information and is intended to carry additional version information,
            for example the version information of the caller.
        token: Long-lived authentication token in plain text format. Used by IQM Resonance.
            If ``token`` is given no other user authentication parameters should be given.
        tokens_file: Path to a tokens file used for authentication.
            If ``tokens_file`` is given no other user authentication parameters should be given.
        auth_server_url: Base URL of the authentication server.
            If ``auth_server_url`` is given also ``username`` and ``password`` must be given.
        username: Username to log in to authentication server.
        password: Password to log in to authentication server.
        api_variant: API variant to use. Default is ``APIVariant.V1``.
            Configurable also by environment variable ``IQM_CLIENT_API_VARIANT``.

    Alternatively, the user authentication related keyword arguments can also be given in
    environment variables :envvar:`IQM_TOKEN`, :envvar:`IQM_TOKENS_FILE`, :envvar:`IQM_AUTH_SERVER`,
    :envvar:`IQM_AUTH_USERNAME` and :envvar:`IQM_AUTH_PASSWORD`. All parameters must be given either
    as keyword arguments or as environment variables. Same combination restrictions apply
    for values given as environment variables as for keyword arguments.

    """

    def __init__(
        self,
        url: str,
        *,
        client_signature: str | None = None,
        token: str | None = None,
        tokens_file: str | None = None,
        auth_server_url: str | None = None,
        username: str | None = None,
        password: str | None = None,
        api_variant: APIVariant | None = None,
    ):
        if not url.startswith(("http:", "https:")):
            raise ClientConfigurationError(f"The URL schema has to be http or https. Incorrect schema in URL: {url}")
        self._token_manager = TokenManager(
            token,
            tokens_file,
            auth_server_url,
            username,
            password,
        )
        version_string = "iqm-client"
        self._signature = f"{platform.platform(terse=True)}"
        self._signature += f", python {platform.python_version()}"
        self._signature += f", iqm-client {version(version_string)}"
        if client_signature:
            self._signature += f", {client_signature}"
        self._architecture: QuantumArchitectureSpecification | None = None
        self._static_architecture: StaticQuantumArchitecture | None = None
        self._dynamic_architectures: dict[UUID, DynamicQuantumArchitecture] = {}

        if api_variant is None:
            env_var = os.environ.get("IQM_CLIENT_API_VARIANT")
            api_variant = APIVariant(env_var) if env_var else APIVariant.V1
        self._api = APIConfig(api_variant, url)
        if (version_incompatibility_msg := self._check_versions()) is not None:
            warnings.warn(version_incompatibility_msg)

    def __del__(self):
        try:
            # try our best to close the auth session, doesn't matter if it fails,
            self.close_auth_session()
        except Exception:
            pass

    def _retry_request_on_error(self, request: Callable[[], requests.Response]) -> requests.Response:
        """Temporary workaround for 502 errors.

        The current implementation of the server side can run out of network connections
        and silently drop incoming connections making IQM Client to fail with 502 errors.
        """
        while True:
            result = request()
            if result.status_code == 502:
                time.sleep(SECONDS_BETWEEN_CALLS)
                continue
            break

        return result

    def _get_request(
        self,
        api_endpoint: APIEndpoint,
        endpoint_args: tuple[str, ...] = (),
        *,
        timeout: float,
        retry: bool = False,
    ) -> requests.Response:
        """Make a HTTP GET request to an IQM server endpoint.

        Contains all the boilerplate code for making a simple GET request.

        Args:
            api_endpoint: API endpoint to GET.
            endpoint_args: Arguments for the endpoint.
            timeout: HTTP request timeout (in seconds).
            retry: Iff True, keep trying if you get a 502 error.

        Returns:
            HTTP response to the request.

        Raises:
            ClientAuthenticationError: No valid authentication provided.
            HTTPError: Various HTTP exceptions.

        """
        url = self._api.url(api_endpoint, *endpoint_args)

        def request():
            return requests.get(
                url,
                headers=self._default_headers(),
                timeout=timeout,
            )

        response = self._retry_request_on_error(request) if retry else request()
        self._check_not_found_error(response)
        self._check_authentication_errors(response)
        response.raise_for_status()
        return response

    def _deserialize_response(
        self,
        response: requests.Response,
        model_class: type[T_BaseModel],
    ) -> T_BaseModel:
        """Deserialize a HTTP endpoint response.

        Args:
            response: HTTP response data.
            model_class: Pydantic model to deserialize the data into.

        Returns:
            Deserialized endpoint response.

        Raises:
            EndpointRequestError: Did not understand the endpoint response.

        """
        try:
            model = model_class.model_validate(response.json())
            # TODO this would be faster but MockJsonResponse.text in our unit tests cannot handle UUID
            # model = model_class.model_validate_json(response.text)
        except json.decoder.JSONDecodeError as e:
            raise EndpointRequestError(f"Invalid response: {response.text}, {e!r}") from e
        return model

    def submit_circuits(
        self,
        circuits: CircuitBatch,
        *,
        qubit_mapping: dict[str, str] | None = None,
        custom_settings: dict[str, Any] | None = None,
        calibration_set_id: UUID | None = None,
        shots: int = 1,
        options: CircuitCompilationOptions | None = None,
    ) -> UUID:
        """Submit a batch of quantum circuits for execution on a quantum computer.

        Args:
            circuits: Circuits to be executed.
            qubit_mapping: Mapping of logical qubit names to physical qubit names.
                Can be set to ``None`` if all ``circuits`` already use physical qubit names.
                Note that the ``qubit_mapping`` is used for all ``circuits``.
            custom_settings: Custom settings to override default settings and calibration data.
                Note: This field should always be ``None`` in normal use.
            calibration_set_id: ID of the calibration set to use, or ``None`` to use the current default calibration.
            shots: Number of times ``circuits`` are executed. Must be greater than zero.
            options: Various discrete options for compiling quantum circuits to instruction schedules.

        Returns:
            ID for the created job. This ID is needed to query the job status and the execution results.

        """
        run_request = self.create_run_request(
            circuits=circuits,
            qubit_mapping=qubit_mapping,
            custom_settings=custom_settings,
            calibration_set_id=calibration_set_id,
            shots=shots,
            options=options,
        )
        job_id = self.submit_run_request(run_request)
        return job_id

    def create_run_request(
        self,
        circuits: CircuitBatch,
        *,
        qubit_mapping: dict[str, str] | None = None,
        custom_settings: dict[str, Any] | None = None,
        calibration_set_id: UUID | None = None,
        shots: int = 1,
        options: CircuitCompilationOptions | None = None,
    ) -> RunRequest:
        """Create a run request for executing circuits without sending it to the server.

        This is called in :meth:`submit_circuits` and does not need to be called separately in normal usage.

        Can be used to inspect the run request that would be submitted by :meth:`submit_circuits`, without actually
        submitting it for execution.

        Args:
            circuits: Circuits to be executed.
            qubit_mapping: Mapping of logical qubit names to physical qubit names.
                Can be set to ``None`` if all ``circuits`` already use physical qubit names.
                Note that the ``qubit_mapping`` is used for all ``circuits``.
            custom_settings: Custom settings to override default settings and calibration data.
                Note: This field should always be ``None`` in normal use.
            calibration_set_id: ID of the calibration set to use, or ``None`` to use the current default calibration.
            shots: Number of times ``circuits`` are executed. Must be greater than zero.
            options: Various discrete options for compiling quantum circuits to instruction schedules.

        Returns:
            RunRequest that would be submitted by equivalent call to :meth:`submit_circuits`.

        """
        if shots < 1:
            raise ValueError("Number of shots must be greater than zero.")
        if options is None:
            options = CircuitCompilationOptions()

        for i, circuit in enumerate(circuits):
            try:
                # validate the circuit against the static information in iqm.iqm_client.models._SUPPORTED_OPERATIONS
                validate_circuit(circuit)
            except ValueError as e:
                raise CircuitValidationError(f"The circuit at index {i} failed the validation").with_traceback(
                    e.__traceback__
                )

        architecture = self.get_dynamic_quantum_architecture(calibration_set_id)

        self._validate_qubit_mapping(architecture, circuits, qubit_mapping)
        serialized_qubit_mapping = serialize_qubit_mapping(qubit_mapping) if qubit_mapping else None

        # validate the circuit against the calibration-dependent dynamic quantum architecture
        self._validate_circuit_instructions(
            architecture,
            circuits,
            qubit_mapping,
            validate_moves=options.move_gate_validation,
            must_close_sandwiches=False,
        )
        return RunRequest(
            qubit_mapping=serialized_qubit_mapping,
            circuits=circuits,
            custom_settings=custom_settings,
            calibration_set_id=calibration_set_id,
            shots=shots,
            max_circuit_duration_over_t2=options.max_circuit_duration_over_t2,
            heralding_mode=options.heralding_mode,
            move_validation_mode=options.move_gate_validation,
            move_gate_frame_tracking_mode=options.move_gate_frame_tracking,
            active_reset_cycles=options.active_reset_cycles,
            dd_mode=options.dd_mode,
            dd_strategy=options.dd_strategy,
        )

    def submit_run_request(self, run_request: RunRequest) -> UUID:
        """Submit a run request for execution on a quantum computer.

        This is called in :meth:`submit_circuits` and does not need to be called separately in normal usage.

        Args:
            run_request: Run request to be submitted for execution.

        Returns:
            ID for the created job. This ID is needed to query the job status and the execution results.

        """
        headers = {"Expect": "100-Continue", **self._default_headers()}
        try:
            # check if someone is trying to profile us with OpenTelemetry
            from opentelemetry import propagate

            propagate.inject(headers)
        except ImportError as _:
            # no OpenTelemetry, no problem
            pass

        if os.environ.get("IQM_CLIENT_DEBUG") == "1":
            print(f"\nIQM CLIENT DEBUGGING ENABLED\nSUBMITTING RUN REQUEST:\n{run_request}\n")

        result = self._retry_request_on_error(
            lambda: requests.post(
                self._api.url(APIEndpoint.SUBMIT_JOB),
                json=json.loads(run_request.model_dump_json(exclude_none=True)),
                headers=headers,
                timeout=REQUESTS_TIMEOUT,
            )
        )

        self._check_not_found_error(result)

        if result.status_code == 401:
            raise ClientAuthenticationError(f"Authentication failed: {result.text}")

        if 400 <= result.status_code < 500:
            raise ClientConfigurationError(f"Client configuration error: {result.text}")

        result.raise_for_status()

        try:
            job_id = UUID(result.json()["id"])
            return job_id
        except (json.decoder.JSONDecodeError, KeyError) as e:
            raise CircuitExecutionError(f"Invalid response: {result.text}, {e}") from e

    @staticmethod
    def _validate_qubit_mapping(
        architecture: DynamicQuantumArchitecture,
        circuits: CircuitBatch,
        qubit_mapping: dict[str, str] | None = None,
    ) -> None:
        """Validate the given qubit mapping.

        Args:
          architecture: Quantum architecture to check against.
          circuits: Circuits to be checked.
          qubit_mapping: Mapping of logical qubit names to physical qubit names.
              Can be set to ``None`` if all ``circuits`` already use physical qubit names.
              Note that the ``qubit_mapping`` is used for all ``circuits``.

        Raises:
            CircuitValidationError: There was something wrong with ``circuits``.

        """
        if qubit_mapping is None:
            return

        # check if qubit mapping is injective
        target_qubits = set(qubit_mapping.values())
        if not len(target_qubits) == len(qubit_mapping):
            raise CircuitValidationError("Multiple logical qubits map to the same physical qubit.")

        # check if qubit mapping covers all qubits in the circuits
        for i, circuit in enumerate(circuits):
            diff = circuit.all_qubits() - set(qubit_mapping)
            if diff:
                raise CircuitValidationError(
                    f"The qubits {diff} in circuit '{circuit.name}' at index {i} "
                    f"are not found in the provided qubit mapping."
                )

        # check that each mapped qubit is defined in the quantum architecture
        for _logical, physical in qubit_mapping.items():
            if physical not in architecture.components:
                raise CircuitValidationError(f"Component {physical} not present in dynamic quantum architecture")

    @staticmethod
    def _validate_circuit_instructions(
        architecture: DynamicQuantumArchitecture,
        circuits: CircuitBatch,
        qubit_mapping: dict[str, str] | None = None,
        validate_moves: MoveGateValidationMode = MoveGateValidationMode.STRICT,
        *,
        must_close_sandwiches: bool = True,
    ) -> None:
        """Validate the given circuits against the given quantum architecture.

        Args:
          architecture: Quantum architecture to check against.
          circuits: Circuits to be checked.
          qubit_mapping: Mapping of logical qubit names to physical qubit names.
              Can be set to ``None`` if all ``circuits`` already use physical qubit names.
              Note that the ``qubit_mapping`` is used for all ``circuits``.
          validate_moves: Determines how MOVE gate validation works.
          must_close_sandwiches: Iff True, MOVE sandwiches cannot be left open when the circuit ends.

        Raises:
            CircuitValidationError: validation failed

        """
        for index, circuit in enumerate(circuits):
            measurement_keys: set[str] = set()
            for instr in circuit.instructions:
                IQMClient._validate_instruction(architecture, instr, qubit_mapping)
                # check measurement key uniqueness
                if instr.name in {"measure", "measurement"}:
                    key = instr.args["key"]
                    if key in measurement_keys:
                        raise CircuitValidationError(f"Circuit {index}: {instr!r} has a non-unique measurement key.")
                    measurement_keys.add(key)
            IQMClient._validate_circuit_moves(
                architecture,
                circuit,
                qubit_mapping,
                validate_moves=validate_moves,
                must_close_sandwiches=must_close_sandwiches,
            )

    @staticmethod
    def _validate_instruction(
        architecture: DynamicQuantumArchitecture,
        instruction: Instruction,
        qubit_mapping: dict[str, str] | None = None,
    ) -> None:
        """Validate an instruction against the dynamic quantum quantum architecture.

        Checks that the instruction uses a valid implementation, and targets a valid locus.

        Args:
          architecture: Quantum architecture to check against.
          instruction: Instruction to check.
          qubit_mapping: Mapping of logical qubit names to physical qubit names.
              Can be set to ``None`` if ``instruction`` already uses physical qubit names.

        Raises:
            CircuitValidationError: validation failed

        """
        op_info = _SUPPORTED_OPERATIONS.get(instruction.name)
        if op_info is None:
            raise CircuitValidationError(f"Unknown quantum operation '{instruction.name}'.")

        # apply the qubit mapping if any
        mapped_qubits = tuple(qubit_mapping[q] for q in instruction.qubits) if qubit_mapping else instruction.qubits

        def check_locus_components(allowed_components: Iterable[str], msg: str) -> None:
            """Checks that the instruction locus consists of the allowed components only."""
            for q, mapped_q in zip(instruction.qubits, mapped_qubits):
                if mapped_q not in allowed_components:
                    raise CircuitValidationError(
                        f"{instruction!r}: Component {q} = {mapped_q} {msg}."
                        if qubit_mapping
                        else f"{instruction!r}: Component {q} {msg}."
                    )

        if op_info.no_calibration_needed:
            # all QPU loci are allowed
            check_locus_components(architecture.components, msg="does not exist on the QPU")
            return

        gate_info = architecture.gates.get(instruction.name)
        if gate_info is None:
            raise CircuitValidationError(
                f"Operation '{instruction.name}' is not supported by the dynamic quantum architecture."
            )

        if instruction.implementation is not None:
            # specific implementation requested
            impl_info = gate_info.implementations.get(instruction.implementation)
            if impl_info is None:
                raise CircuitValidationError(
                    f"Operation '{instruction.name}' implementation '{instruction.implementation}' "
                    f"is not supported by the dynamic quantum architecture."
                )
            allowed_loci = impl_info.loci
            instruction_name = f"{instruction.name}.{instruction.implementation}"
        else:
            # any implementation is fine
            allowed_loci = gate_info.loci
            instruction_name = f"{instruction.name}"

        if op_info.factorizable:
            # Check that all the locus components are allowed by the architecture
            check_locus_components(
                {q for locus in allowed_loci for q in locus}, msg=f"is not allowed as locus for '{instruction_name}'"
            )
            return

        # Check that locus matches one of the allowed loci
        all_loci = (
            tuple(tuple(x) for locus in allowed_loci for x in itertools.permutations(locus))
            if op_info.symmetric
            else allowed_loci
        )
        if mapped_qubits not in all_loci:
            raise CircuitValidationError(
                f"{instruction.qubits} = {tuple(mapped_qubits)} is not allowed as locus for '{instruction_name}'"
                if qubit_mapping
                else f"{instruction.qubits} is not allowed as locus for '{instruction_name}'"
            )

    @staticmethod
    def _validate_circuit_moves(
        architecture: DynamicQuantumArchitecture,
        circuit: Circuit,
        qubit_mapping: dict[str, str] | None = None,
        validate_moves: MoveGateValidationMode = MoveGateValidationMode.STRICT,
        *,
        must_close_sandwiches: bool = True,
    ) -> None:
        """Raise an error if the MOVE gates in the circuit are not valid in the given architecture.

        Args:
            architecture: Quantum architecture to check against.
            circuit: Quantum circuit to validate.
            qubit_mapping: Mapping of logical qubit names to physical qubit names.
                Can be set to ``None`` if the ``circuit`` already uses physical qubit names.
            validate_moves: Option for bypassing full or partial MOVE gate validation.
            must_close_sandwiches: Iff True, MOVE sandwiches cannot be left open when the circuit ends.

        Raises:
            CircuitValidationError: validation failed

        """
        if validate_moves == MoveGateValidationMode.NONE:
            return
        move_gate = "move"
        # Check if MOVE gates are allowed on this architecture
        if move_gate not in architecture.gates:
            if any(i.name == move_gate for i in circuit.instructions):
                raise CircuitValidationError("MOVE instruction is not supported by the given device architecture.")
            return

        # some gates are allowed in MOVE sandwiches
        allowed_gates = {"barrier"}
        if validate_moves == MoveGateValidationMode.ALLOW_PRX:
            allowed_gates.add("prx")

        all_resonators = set(architecture.computational_resonators)
        all_qubits = set(architecture.qubits)
        if qubit_mapping:
            reverse_mapping = {phys: log for log, phys in qubit_mapping.items()}
            all_resonators = {reverse_mapping[q] if q in reverse_mapping else q for q in all_resonators}
            all_qubits = {reverse_mapping[q] if q in reverse_mapping else q for q in all_qubits}

        # Mapping from resonator to the qubit whose state it holds. Resonators not in the map hold no qubit state.
        resonator_occupations: dict[str, str] = {}
        # Qubits whose states are currently moved to a resonator
        moved_qubits: set[str] = set()

        for inst in circuit.instructions:
            if inst.name == "move":
                qubit, resonator = inst.qubits
                if not (qubit in all_qubits and resonator in all_resonators):
                    raise CircuitValidationError(
                        f"MOVE instructions are only allowed between qubit and resonator, not {inst.qubits}."
                    )

                if (resonator_qubit := resonator_occupations.get(resonator)) is None:
                    # Beginning MOVE: check that the qubit hasn't been moved to another resonator
                    if qubit in moved_qubits:
                        raise CircuitValidationError(
                            f"MOVE instruction {inst.qubits}: state of {qubit} is "
                            f"in another resonator: {resonator_occupations}."
                        )
                    resonator_occupations[resonator] = qubit
                    moved_qubits.add(qubit)
                else:
                    # Ending MOVE: check that the qubit matches to the qubit that was moved to the resonator
                    if resonator_qubit != qubit:
                        raise CircuitValidationError(
                            f"MOVE instruction {inst.qubits} to an already occupied resonator: {resonator_occupations}."
                        )
                    del resonator_occupations[resonator]
                    moved_qubits.remove(qubit)
            elif moved_qubits:
                # Validate that moved qubits are not used during MOVE operations
                if inst.name not in allowed_gates:
                    if overlap := set(inst.qubits) & moved_qubits:
                        raise CircuitValidationError(
                            f"Instruction {inst.name} acts on {inst.qubits} while the state(s) of {overlap} "
                            f"are in a resonator. Current resonator occupation: {resonator_occupations}."
                        )

        # Finally validate that all MOVE sandwiches have been ended before the circuit ends
        if must_close_sandwiches and resonator_occupations:
            raise CircuitValidationError(
                f"Circuit ends while qubit state(s) are still in a resonator: {resonator_occupations}."
            )

    def _get_run_v1(self, job_id: UUID, timeout_secs: float = REQUESTS_TIMEOUT) -> RunResult:
        """V1 API (Cocos circuit execution and Resonance) has an inefficient `GET /jobs/<id>` endpoint
        that returns the full job status, including the result and the original request, in a single call.
        """
        response = self._get_request(
            APIEndpoint.GET_JOB_RESULT,
            (str(job_id),),
            timeout=timeout_secs,
            retry=True,
        )
        return self._deserialize_response(response, RunResult)

    def _get_run_v2(self, job_id: UUID, timeout_secs: float = REQUESTS_TIMEOUT) -> RunResult:
        """V2 API (Station-based circuit execution) has granular endpoints for job status and result."""
        status_response = self._get_request(
            APIEndpoint.GET_JOB_STATUS,
            (str(job_id),),
            timeout=timeout_secs,
        )
        status = status_response.json()
        if Status(status["status"]) not in Status.terminal_statuses():
            return RunResult.from_dict(
                {
                    "measurements": [],
                    "status": status["status"],
                    "message": "",
                    "metadata": {
                        "calibration_set_id": None,
                        "circuits_batch": [],
                        "parameters": None,
                        "timestamps": {},
                    },
                }
            )

        result = self._retry_request_on_error(
            lambda: requests.get(
                self._api.url(APIEndpoint.GET_JOB_RESULT, str(job_id)),
                headers=self._default_headers(),
                timeout=timeout_secs,
            )
        )
        if result.status_code != 404:
            result.raise_for_status()
        measurements = [] if result.status_code == 404 else result.json()
        request_parameters = (
            {}
            if result.status_code == 404
            else requests.get(
                self._api.url(APIEndpoint.GET_JOB_REQUEST_PARAMETERS, str(job_id)),
                headers=self._default_headers(),
                timeout=timeout_secs,
            ).json()
        )
        calibration_set_id = (
            None
            if result.status_code == 404
            else requests.get(
                self._api.url(APIEndpoint.GET_JOB_CALIBRATION_SET_ID, str(job_id)),
                headers=self._default_headers(),
                timeout=timeout_secs,
            ).json()
        )
        circuits_batch = (
            []
            if result.status_code == 404
            else requests.get(
                self._api.url(APIEndpoint.GET_JOB_CIRCUITS_BATCH, str(job_id)),
                headers=self._default_headers(),
                timeout=timeout_secs,
            ).json()
        )
        timeline = (
            []
            if result.status_code == 404
            else requests.get(
                self._api.url(APIEndpoint.GET_JOB_TIMELINE, str(job_id)),
                headers=self._default_headers(),
                timeout=timeout_secs,
            ).json()
        )
        error_message_response = requests.get(
            self._api.url(APIEndpoint.GET_JOB_ERROR_LOG, str(job_id)),
            headers=self._default_headers(),
            timeout=timeout_secs,
        )
        error_message = error_message_response.text if error_message_response.status_code == 200 else None
        return RunResult.from_dict(
            {
                "measurements": measurements,
                "status": status["status"],
                "message": error_message,
                "metadata": {
                    "`": calibration_set_id,
                    "circuits_batch": circuits_batch,
                    "parameters": (
                        None
                        if result.status_code == 404
                        else {
                            "shots": request_parameters["shots"],
                            "max_circuit_duration_over_t2": request_parameters["max_circuit_duration_over_t2"],
                            "heralding_mode": request_parameters["heralding_mode"],
                            "move_validation_mode": request_parameters["move_validation_mode"],
                            "move_gate_frame_tracking_mode": request_parameters["move_gate_frame_tracking_mode"],
                        }
                    ),
                    "timestamps": {datapoint["stage"]: datapoint["timestamp"] for datapoint in timeline},
                },
            }
        )

    def get_run(self, job_id: UUID, *, timeout_secs: float = REQUESTS_TIMEOUT) -> RunResult:
        """Query the status and results of a submitted job.

        Args:
            job_id: ID of the job to query.
            timeout_secs: Network request timeout (seconds).

        Returns:
            Result of the job (can be pending).

        Raises:
            CircuitExecutionError: IQM server specific exceptions
            HTTPException: HTTP exceptions

        """
        if self._api.variant == APIVariant.V2:
            run_result = self._get_run_v2(job_id, timeout_secs)
        else:
            run_result = self._get_run_v1(job_id, timeout_secs)

        if run_result.warnings:
            for warning in run_result.warnings:
                warnings.warn(warning)
        if run_result.status == Status.FAILED:
            raise CircuitExecutionError(run_result.message)
        return run_result

    def get_run_status(self, job_id: UUID, *, timeout_secs: float = REQUESTS_TIMEOUT) -> RunStatus:
        """Query the status of a submitted job.

        Args:
            job_id: ID of the job to query.
            timeout_secs: Network request timeout (seconds).

        Returns:
            Job status.

        Raises:
            CircuitExecutionError: IQM server specific exceptions
            HTTPException: HTTP exceptions

        """
        response = self._get_request(
            APIEndpoint.GET_JOB_STATUS,
            (str(job_id),),
            timeout=timeout_secs,
            retry=True,
        )
        run_status = self._deserialize_response(response, RunStatus)

        if run_status.warnings:
            for warning in run_status.warnings:
                warnings.warn(warning)
        return run_status

    def wait_for_compilation(self, job_id: UUID, timeout_secs: float = DEFAULT_TIMEOUT_SECONDS) -> RunResult:
        """Poll results until a job is either compiled, pending execution, ready, failed, aborted, or timed out.

        Args:
            job_id: ID of the job to wait for.
            timeout_secs: How long to wait for a response before raising an APITimeoutError (seconds).

        Returns:
            Job result.

        Raises:
            APITimeoutError: time exceeded the set timeout

        """
        start_time = datetime.now()
        while (datetime.now() - start_time).total_seconds() < timeout_secs:
            status = self.get_run_status(job_id).status
            if status in Status.terminal_statuses() | {Status.PENDING_EXECUTION, Status.COMPILED}:
                return self.get_run(job_id)
            time.sleep(SECONDS_BETWEEN_CALLS)
        raise APITimeoutError(f"The job compilation didn't finish in {timeout_secs} seconds.")

    def wait_for_results(self, job_id: UUID, timeout_secs: float = DEFAULT_TIMEOUT_SECONDS) -> RunResult:
        """Poll results until a job is either ready, failed, aborted, or timed out.

           Note that jobs handling on the server side is async and if we try to request the results
           right after submitting the job (which is usually the case)
           we will find the job is still pending at least for the first query.

        Args:
            job_id: ID of the job to wait for.
            timeout_secs: How long to wait for a response before raising an APITimeoutError (seconds).

        Returns:
            Job result.

        Raises:
            APITimeoutError: time exceeded the set timeout

        """
        start_time = datetime.now()
        while (datetime.now() - start_time).total_seconds() < timeout_secs:
            status = self.get_run_status(job_id).status
            if status in Status.terminal_statuses():
                return self.get_run(job_id)
            time.sleep(SECONDS_BETWEEN_CALLS)
        raise APITimeoutError(f"The job didn't finish in {timeout_secs} seconds.")

    def abort_job(self, job_id: UUID, *, timeout_secs: float = REQUESTS_TIMEOUT) -> None:
        """Abort a job that was submitted for execution.

        Args:
            job_id: ID of the job to be aborted.
            timeout_secs: Network request timeout (seconds).

        Raises:
            JobAbortionError: aborting the job failed

        """
        result = requests.post(
            self._api.url(APIEndpoint.ABORT_JOB, str(job_id)),
            headers=self._default_headers(),
            timeout=timeout_secs,
        )
        if result.status_code != 200:
            raise JobAbortionError(result.text)

    def get_quantum_architecture(self, *, timeout_secs: float = REQUESTS_TIMEOUT) -> QuantumArchitectureSpecification:
        """Retrieve quantum architecture from server.

        Caches the result and returns it on later invocations.

        Args:
            timeout_secs: Network request timeout (seconds).

        Returns:
            Quantum architecture of the server.

        Raises:
            EndpointRequestError: did not understand the endpoint response
            ClientAuthenticationError: no valid authentication provided
            HTTPException: HTTP exceptions

        """
        if self._architecture:
            return self._architecture

        response = self._get_request(
            APIEndpoint.QUANTUM_ARCHITECTURE,
            timeout=timeout_secs,
        )
        qa = self._deserialize_response(response, QuantumArchitecture).quantum_architecture

        # Cache architecture so that later invocations do not need to query it again
        self._architecture = qa
        return qa

    def get_static_quantum_architecture(self, *, timeout_secs: float = REQUESTS_TIMEOUT) -> StaticQuantumArchitecture:
        """Retrieve the static quantum architecture (SQA) from the server.

        Caches the result and returns it on later invocations.

        Args:
            timeout_secs: Network request timeout (seconds).

        Returns:
            Static quantum architecture of the server.

        Raises:
            EndpointRequestError: did not understand the endpoint response
            ClientAuthenticationError: no valid authentication provided
            HTTPException: HTTP exceptions

        """
        if self._static_architecture:
            return self._static_architecture

        response = self._get_request(
            APIEndpoint.STATIC_QUANTUM_ARCHITECTURE,
            timeout=timeout_secs,
        )
        sqa = self._deserialize_response(response, StaticQuantumArchitecture)

        # Cache the architecture so that later invocations do not need to query it again
        self._static_architecture = sqa
        return sqa

    def get_quality_metric_set(
        self, calibration_set_id: UUID | None = None, *, timeout_secs: float = REQUESTS_TIMEOUT
    ) -> QualityMetricSet:
        """Retrieve the latest quality metric set for the given calibration set from the server.

        Args:
            calibration_set_id: ID of the calibration set for which the quality metrics are returned.
                If ``None``, the current default calibration set is used.
            timeout_secs: Network request timeout (seconds).

        Returns:
            Requested quality metric set.

        Raises:
            EndpointRequestError: did not understand the endpoint response
            ClientAuthenticationError: no valid authentication provided
            HTTPException: HTTP exceptions

        """
        if calibration_set_id is None:
            calibration_set_id_str = "default"
        else:
            calibration_set_id_str = str(calibration_set_id)

        response = self._get_request(
            APIEndpoint.QUALITY_METRICS,
            (calibration_set_id_str,),
            timeout=timeout_secs,
        )
        return self._deserialize_response(response, QualityMetricSet)

    def get_calibration_set(
        self, calibration_set_id: UUID | None = None, *, timeout_secs: float = REQUESTS_TIMEOUT
    ) -> CalibrationSet:
        """Retrieve the given calibration set from the server.

        Args:
            calibration_set_id: ID of the calibration set to retrieve.
                If ``None``, the current default calibration set is retrieved.
            timeout_secs: Network request timeout (seconds).

        Returns:
            Requested calibration set.

        Raises:
            EndpointRequestError: did not understand the endpoint response
            ClientAuthenticationError: no valid authentication provided
            HTTPException: HTTP exceptions

        """
        if calibration_set_id is None:
            calibration_set_id_str = "default"
        else:
            calibration_set_id_str = str(calibration_set_id)

        response = self._get_request(
            APIEndpoint.CALIBRATION,
            (calibration_set_id_str,),
            timeout=timeout_secs,
        )
        return self._deserialize_response(response, CalibrationSet)

    def get_dynamic_quantum_architecture(
        self, calibration_set_id: UUID | None = None, *, timeout_secs: float = REQUESTS_TIMEOUT
    ) -> DynamicQuantumArchitecture:
        """Retrieve the dynamic quantum architecture (DQA) for the given calibration set from the server.

        Caches the result and returns the same result on later invocations, unless ``calibration_set_id`` is ``None``.
        If ``calibration_set_id`` is ``None``, always retrieves the result from the server because the default
        calibration set may have changed.

        Args:
            calibration_set_id: ID of the calibration set for which the DQA is retrieved.
                If ``None``, use current default calibration set on the server.
            timeout_secs: Network request timeout (seconds).

        Returns:
            Dynamic quantum architecture corresponding to the given calibration set.

        Raises:
            EndpointRequestError: did not understand the endpoint response
            ClientAuthenticationError: no valid authentication provided
            HTTPException: HTTP exceptions

        """
        if calibration_set_id is None:
            calibration_set_id_str = "default"
        elif calibration_set_id in self._dynamic_architectures:
            return self._dynamic_architectures[calibration_set_id]
        else:
            calibration_set_id_str = str(calibration_set_id)

        response = self._get_request(
            APIEndpoint.CALIBRATED_GATES,
            (calibration_set_id_str,),
            timeout=timeout_secs,
        )
        dqa = self._deserialize_response(response, DynamicQuantumArchitecture)

        # Cache architecture so that later invocations do not need to query it again
        self._dynamic_architectures[dqa.calibration_set_id] = dqa
        return dqa

    def get_feedback_groups(self, *, timeout_secs: float = REQUESTS_TIMEOUT) -> tuple[frozenset[str], ...]:
        """Retrieve groups of qubits that can receive real-time feedback signals from each other.

        Real-time feedback enables conditional gates such as `cc_prx`.
        Some hardware configurations support routing real-time feedback only between certain qubits.

        This method is only supported for the API variant V2.

        Args:
            timeout_secs: Network request timeout (seconds).

        Returns:
            Feedback groups. Within a group, any qubit can receive real-time feedback from any other qubit in
                the same group. A qubit can belong to multiple groups.
                If there is only one group, there are no restrictions regarding feedback routing.

        Raises:
            EndpointRequestError: did not understand the endpoint response
            ClientAuthenticationError: no valid authentication provided
            HTTPException: HTTP exceptions

        """
        response = self._get_request(
            APIEndpoint.CHANNEL_PROPERTIES,
            timeout=timeout_secs,
        )
        try:
            channel_properties = DictDict.validate_json(response.text)
        except PydanticValidationError as e:
            raise EndpointRequestError(f"Invalid response: {response.text}, {e!r}") from e

        all_qubits = self.get_quantum_architecture().qubits
        groups: dict[str, set[str]] = {}
        # All qubits that can read from the same source belong to the same group.
        # A qubit may belong to multiple groups.
        for channel_name, properties in channel_properties.items():
            # Relying on naming convention because we don't have proper mapping available:
            qubit = channel_name.split("__")[0]
            if qubit not in all_qubits:
                continue
            for source in properties.get("fast_feedback_sources", ()):
                groups.setdefault(source, set()).add(qubit)
        # Merge identical groups
        unique_groups: set[frozenset[str]] = {frozenset(group) for group in groups.values()}
        # Sort by group size
        return tuple(sorted(unique_groups, key=len, reverse=True))

    def close_auth_session(self) -> bool:
        """Terminate session with authentication server if there is one.

        Returns:
            True iff session was successfully closed.

        Raises:
            ClientAuthenticationError: logout failed
            ClientAuthenticationError: asked to close externally managed authentication session

        """
        return self._token_manager.close()

    def _default_headers(self) -> dict[str, str]:
        """Default headers for HTTP requests to the IQM server."""
        headers = {"User-Agent": self._signature}
        if bearer_token := self._token_manager.get_bearer_token():
            headers["Authorization"] = bearer_token
        return headers

    @staticmethod
    def _check_authentication_errors(result: requests.Response) -> None:
        """Raise ClientAuthenticationError with appropriate message if the authentication failed for some reason."""
        # for not strictly authenticated endpoints,
        # we need to handle 302 redirects to the auth server login page
        if result.history and any(
            response.status_code == 302 for response in result.history
        ):  # pragma: no cover (generators are broken in coverage)
            raise ClientAuthenticationError("Authentication is required.")
        if result.status_code == 401:
            raise ClientAuthenticationError(f"Authentication failed: {result.text}")

    def _check_not_found_error(self, response: requests.Response) -> None:
        """Raise HTTPError with appropriate message if ``response.status_code == 404``."""
        if response.status_code == 404:
            version_message = ""
            if (version_incompatibility_msg := self._check_versions()) is not None:
                version_message = (
                    f" This may be caused by the server version not supporting this endpoint. "
                    f"{version_incompatibility_msg}"
                )
            raise HTTPError(f"{response.url} not found.{version_message}", response=response)

    def _check_versions(self) -> str | None:
        """Check the client version against compatible client versions reported by server.

        Returns:
            Message about client incompatibility with the server if the versions are incompatible or if the
            compatibility could not be confirmed, ``None`` if they are compatible.

        """
        try:
            libraries = self.get_supported_client_libraries()
            compatible_iqm_client = libraries.get(
                "iqm-client",
                libraries.get("iqm_client"),
            )
            if compatible_iqm_client is None:
                return "Could not verify IQM Client compatibility with the server. You might encounter issues."
            min_version = parse(compatible_iqm_client.min)
            max_version = parse(compatible_iqm_client.max)
            client_version = parse(version("iqm-client"))
            if client_version < min_version or client_version >= max_version:
                return (
                    f"Your IQM Client version {client_version} was built for a different version of IQM Server. "
                    f"You might encounter issues. For the best experience, consider using a version "
                    f"of IQM Client that satisfies {min_version} <= iqm-client < {max_version}."
                )
            return None
        except Exception as e:
            # we don't want the version check to prevent usage of IQMClient in any situation
            check_error = e
        return f"Could not verify IQM Client compatibility with the server. You might encounter issues. {check_error}"

    def get_run_counts(self, job_id: UUID, *, timeout_secs: float = REQUESTS_TIMEOUT) -> RunCounts:
        """Query the counts of an executed job.

        Args:
            job_id: ID of the job to query.
            timeout_secs: Network request timeout (seconds).

        Returns:
            Measurement results of the job in histogram representation.

        Raises:
            EndpointRequestError: did not understand the endpoint response
            ClientAuthenticationError: no valid authentication provided
            HTTPException: HTTP exceptions

        """
        response = self._get_request(
            APIEndpoint.GET_JOB_COUNTS,
            (str(job_id),),
            timeout=timeout_secs,
            retry=True,
        )
        return self._deserialize_response(response, RunCounts)

    def get_supported_client_libraries(self, timeout_secs: float = REQUESTS_TIMEOUT) -> dict[str, ClientLibrary]:
        """Retrieve information about supported client libraries from the server.

        Args:
            timeout_secs: Network request timeout (seconds).

        Returns:
            Mapping from library identifiers to their metadata.

        Raises:
            EndpointRequestError: did not understand the endpoint response
            ClientAuthenticationError: no valid authentication provided
            HTTPException: HTTP exceptions

        """
        response = self._get_request(
            APIEndpoint.CLIENT_LIBRARIES,
            timeout=timeout_secs,
        )
        try:
            return ClientLibraryDict.validate_json(response.text)
        except PydanticValidationError as e:
            raise EndpointRequestError(f"Invalid response: {response.text}, {e!r}") from e



# Copyright 2024 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file is an example of using Qiskit on IQM to run a simple but non-trivial quantum circuit on
Resonance, the IQM quantum cloud service.
See the Qiskit on IQM user guide for instructions:
https://docs.meetiqm.com/iqm-client/user_guide_qiskit.html
"""

import argparse

from iqm.qiskit_iqm import IQMProvider
from qiskit import QuantumCircuit, transpile


def resonance_example(server_url: str, api_token: str | None) -> dict[str, int]:
    """Run a circuit via IQM Resonance.

    Args:
        server_url: URL of the IQM Resonance QC used for execution
        api_token: IQM Resonance API token for authentication

    Returns:
        a mapping of bitstrings representing qubit measurement results to counts for each result

    """
    SHOTS = 1000

    # Initialize a backend
    backend = IQMProvider(server_url, token=api_token).get_backend()

    # Define a quantum circuit
    num_qb = min(backend.num_qubits, 5)  # use at most 5 qubits
    qc = QuantumCircuit(num_qb)

    qc.h(0)
    for qb in range(1, num_qb):
        qc.cx(0, qb)
    qc.barrier()
    qc.measure_all()

    # Transpile the circuit
    qc_transpiled = transpile(qc, backend)
    print(qc_transpiled.draw(output="text"))

    # Run the circuit
    job = backend.run(qc_transpiled, shots=SHOTS)
    return job.result().get_counts()


if __name__ == "__main__":
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "--url",
        help="URL of the IQM Resonance QC",
        # For example https://cocos.resonance.meetiqm.com/garnet
        default="https://cocos.resonance.meetiqm.com/<QUANTUM COMPUTER>",
    )
    argparser.add_argument(
        "--token",
        help="IQM Resonance access token",
        # Provide the API token explicitly or set it as an environment variable
        # following the https://docs.meetiqm.com/iqm-client/user_guide_qiskit.html#authentication
        default="<INSERT YOUR TOKEN>",
    )

    args = argparser.parse_args()
    results = resonance_example(args.url, args.token)
    print(results)



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Demonstrates transforming a quantum circuit into the native gateset and connectivity of a given
IQM device, optimizing it, and then executing it on a simulator.
"""

from __future__ import annotations

import cirq
import cirq.linalg
from iqm.cirq_iqm import IQMDevice
from iqm.cirq_iqm.optimizers import simplify_circuit
import numpy as np

np.set_printoptions(precision=3)


def get_measurements(circuit: cirq.Circuit) -> list[tuple[int, cirq.GateOperation]]:
    """Find all measurements in a circuit.

    Args:
        circuit: quantum circuit

    Returns:
        tuples of (moment index, measurement operation)

    """
    if not circuit.are_all_measurements_terminal():
        raise ValueError("Non-terminal measurements are not supported")

    return [(m[0], m[1]) for m in circuit.findall_operations_with_gate_type(cirq.MeasurementGate)]


def get_qubit_order(circuit: cirq.Circuit) -> tuple[list[cirq.Qid], int]:
    """Ordering for the qubits based on lexical ordering of measurements in the circuit.

    Args:
        circuit: quantum circuit

    Returns:
        qubit order, number of measured qubits

    """
    measurements = get_measurements(circuit)

    # mapping from (key, index) to measured qubit, in lexical order
    m_qubits = {(op.gate.key, index): qubit for _, op in measurements for index, qubit in enumerate(op.qubits)}
    m_qubits = dict(sorted(m_qubits.items()))

    qubit_order = list(m_qubits.values())
    assert len(set(qubit_order)) == len(qubit_order)  # no qubit measured more than once

    # add unobserved qubits to the order
    all_qubits = set(circuit.all_qubits())
    qubit_order += sorted(all_qubits - set(qubit_order))
    return qubit_order, len(m_qubits)


def simulate_measurement_probabilities(
    sim: cirq.Simulator,
    circuit: cirq.Circuit,
) -> tuple[dict[tuple, float], list[str]]:
    """Simulate the probabilities of obtaining different measurement results in the given circuit.

    Args:
        sim: simulator
        circuit: quantum circuit

    Returns:
        map from result to its probability, measurement keys

    """
    measurements = get_measurements(circuit)
    if not measurements:
        raise ValueError("Circuit has no measurements.")

    qubit_order, n_measured_qubits = get_qubit_order(circuit)

    stripped = circuit.copy()
    stripped.batch_remove(measurements)

    # compute the state vector
    result = sim.simulate(stripped, qubit_order=qubit_order)
    state = result.state_vector()

    # trace out non-measured qubits, compute the probabilities of the various measurement outcomes
    mixture = cirq.linalg.partial_trace_of_state_vector_as_mixture(state, keep_indices=range(n_measured_qubits))
    _temp = [p * np.abs(ket) ** 2 for p, ket in mixture]
    probs = np.sum(_temp, axis=0)

    # list the all the measurement outcomes in the matching order
    measurement_arities = {op.gate.key: len(op.qubits) for _, op in measurements}
    measurement_arities = dict(sorted(measurement_arities.items()))

    shape = np.array(2) ** list(measurement_arities.values())
    outcomes = [tuple(seq) for seq in np.array(np.unravel_index(range(len(probs)), shape)).T]
    assert len(outcomes) == len(probs)

    probabilities = dict(zip(outcomes, probs))
    print("    ", probabilities)
    return probabilities, list(measurement_arities.keys())


def simulate_without_measurements(
    sim: cirq.Simulator,
    circuit: cirq.Circuit,
) -> np.ndarray:
    """Simulates given quantum circuit without its measurement operations.

    Args:
        sim: simulator
        circuit: quantum circuit

    Returns:
        state vector

    """
    measurements = get_measurements(circuit)
    stripped = circuit.copy()
    stripped.batch_remove(measurements)

    qubit_order, _ = get_qubit_order(circuit)

    result = sim.simulate(stripped, qubit_order=qubit_order)
    state = result.state_vector()
    print("    probabilities =", np.abs(state) ** 2)
    return state


def pause() -> None:
    input("\npress enter\n")


def demo(device: IQMDevice, circuit: cirq.Circuit, *, use_qsim: bool = False) -> None:
    """Transform the given circuit to a form the given device accepts, then simulate it.

    Args:
        device: device on which to execute the quantum circuit
        circuit: quantum circuit
        use_qsim: Iff True, use the ``qsim`` circuit simulator instead of the Cirq builtin simulator.

    """
    print("Source circuit:")
    print(circuit)
    pause()

    # decompose non-native gates and simplify the circuit
    circuit_decomposed = device.decompose_circuit(circuit)
    circuit_simplified = simplify_circuit(circuit_decomposed)  # NOTE: will discard Rz:s before measurements
    print("\nSimplified circuit:")
    print(circuit_simplified)
    pause()

    # map the circuit qubits to device qubits
    circuit_mapped, initial_mapping, final_mapping = device.route_circuit(circuit_simplified)

    print("\nRouted simplified circuit:")
    print(circuit_mapped)
    pause()

    # Decompose any non-native gates which might have been
    # introduced since the last decomposition.
    circuit_transformed = device.decompose_circuit(circuit_mapped)
    print("\nFinal transformed circuit:")
    print(circuit_transformed)
    pause()

    # Initialize a ket-based simulator for evaluating the circuit
    if use_qsim:
        import qsimcirq

        sim = qsimcirq.QSimSimulator()
        print("Using qsim.")
    else:
        sim = cirq.Simulator()

    # sim.run only returns the measured values over a number of repetitions of the circuit, like a hardware QPU would
    # sim.simulate returns also the simulated state of the register

    print("\nState vector simulation")
    print("\n  Original circuit:")
    state_0 = simulate_without_measurements(sim, circuit)
    print("\n  Transformed circuit:")
    state_1 = simulate_without_measurements(sim, circuit_transformed)

    # Overlap won't be perfect due to use of simplify_circuit above, which may eliminate Rz gates
    # overlap = np.abs(np.vdot(state_0, state_1))
    # print('\noverlap = |<original|transformed>| =', overlap)
    # assert np.abs(overlap - 1.0) < 1e-6, 'Circuits are not equivalent!'

    # Basis state probabilities, however, are preserved.
    probs_0 = np.abs(state_0) ** 2
    probs_1 = np.abs(state_1) ** 2
    assert np.allclose(probs_0, probs_1), "Circuits are not equivalent!"

    print("\nSimulated measurement probabilities:")
    print("\n  Original circuit:")
    m_probs0, keys0 = simulate_measurement_probabilities(sim, circuit)
    print("\n  Transformed circuit:")
    m_probs1, keys1 = simulate_measurement_probabilities(sim, circuit_transformed)

    assert keys0 == keys1, "Key order is not the same!"
    print(f"\nOrder of measurement keys: {keys0}")

    # Simulated measurement probabilities are also preserved,
    assert np.allclose(list(m_probs0.values()), list(m_probs1.values())), "Circuits are not equivalent!"

    # and they are equivalent with the basis state probabilities when every qubit is measured.
    if len(m_probs0) == len(probs_0):
        assert np.allclose(list(m_probs0.values()), probs_0), "Simulation methods are not equivalent!"

    # Random sampling of the circuit returns data that follows the measurement probabilities.
    samples_0 = sim.run(circuit, repetitions=10000)
    samples_1 = sim.run(circuit_transformed, repetitions=10000)
    # compute sorted histograms of the samples, where order of the bits in the bitstring is determined
    # by the order of the corresponding measurement keys
    hist_0 = dict(sorted(samples_0.multi_measurement_histogram(keys=keys0).items()))
    hist_1 = dict(sorted(samples_1.multi_measurement_histogram(keys=keys0).items()))

    print("\nRandom sampling:")
    print("\n  Original circuit:")
    print("    ", hist_0)
    print("\n  Transformed circuit:")
    print("    ", hist_1)



# Copyright 2024 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A simple extension of the QuantumCircuit class to allow the MOVE
gate to be applied with a .move(qubit, resonator) shortcut.
"""

from iqm.qiskit_iqm.move_gate import MoveGate
from qiskit import QuantumCircuit


class IQMCircuit(QuantumCircuit):
    """Extends the QuantumCircuit class, adding a shortcut for applying the MOVE gate."""

    def move(self, qubit: int, resonator: int):
        """Applies the MOVE gate to the circuit.

        Note: at this point the circuit layout is only guaranteed to work if the order
        of the qubit and the resonator is correct (qubit first, resonator second).

        Args:
            qubit: the logical index of the qubit
            resonator: the logical index of the resonator

        """
        self.append(MoveGate(), [qubit, resonator])



# Copyright 2020–2021 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Imports OpenQASM 2.0 programs in a way that preserves gates native to the IQM architectures."""

import cirq
from cirq.contrib.qasm_import._parser import QasmGateStatement, QasmParser, QasmUGate
import numpy as np


def circuit_from_qasm(qasm: str) -> cirq.circuits.Circuit:
    """Parses an OpenQASM 2.0 program to a Cirq circuit.

    Args:
        qasm: OpenQASM program

    Returns:
        parsed circuit

    """
    parser = QasmParser()

    def convert_U(args: list[float]) -> cirq.Gate:
        """Maps the OpenQASM builtin one-qubit gate ``U`` to :class:`cirq.QasmUGate`,
        or :class:`cirq.PhasedXPowGate` if possible.
        """
        theta, phi, lmda = args
        scale = np.pi
        if phi == -lmda:
            return cirq.PhasedXPowGate(exponent=theta / scale, phase_exponent=phi / scale + 0.5, global_shift=-0.5)
        return QasmUGate(*(p / np.pi for p in args))

    rule = QasmGateStatement(
        qasm_gate="U",
        num_params=3,
        num_args=1,
        cirq_gate=convert_U,
    )
    parser.basic_gates["U"] = rule
    parser.all_gates["U"] = rule
    # u3 is not a basic OpenQASM gate
    parser.all_gates["u3"] = QasmGateStatement(
        qasm_gate="u3",
        num_params=3,
        num_args=1,
        cirq_gate=convert_U,
    )
    return parser.parse(qasm).circuit



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Circuit sampler that executes quantum circuits on an IQM quantum computer."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
import sys
from uuid import UUID
import warnings

import cirq
from iqm.cirq_iqm.devices.iqm_device import IQMDevice, IQMDeviceMetadata
from iqm.cirq_iqm.serialize import serialize_circuit
from iqm.iqm_client import CircuitCompilationOptions, IQMClient, JobAbortionError, RunRequest
import numpy as np


class IQMSampler(cirq.work.Sampler):
    """Circuit sampler for executing quantum circuits on IQM quantum computers.

    IQMSampler connects to a quantum computer through an IQM server.
    If the server requires user authentication, you can provide it either using environment
    variables, or as keyword arguments to IQMSampler. The user authentication kwargs are passed
    through to :class:`~iqm.iqm_client.iqm_client.IQMClient` as is, and are documented there.

    Args:
        url: URL of the IQM server. Has to start with http or https.
        device: Device to execute the circuits on. If ``None``, the device will be created based
            on the calibration-specific dynamic quantum architecture obtained from
            :class:`~iqm.iqm_client.iqm_client.IQMClient`.
        calibration_set_id:
            ID of the calibration set to use. If ``None``, use the default one.
        run_sweep_timeout:
            Timeout for polling sweep results, in seconds. If ``None``, use the client default value.
        compiler_options: The compilation options to use for the circuits, as defined by IQM Client.

    """

    def __init__(
        self,
        url: str,
        *,
        device: IQMDevice | None = None,
        calibration_set_id: UUID | None = None,
        run_sweep_timeout: int | None = None,
        compiler_options: CircuitCompilationOptions | None = None,
        **user_auth_args,  # contains keyword args auth_server_url, username and password
    ):
        self._client = IQMClient(url, **user_auth_args)
        dqa = self._client.get_dynamic_quantum_architecture(calibration_set_id)
        server_device_metadata = IQMDeviceMetadata.from_architecture(dqa)
        self._use_default_calibration_set = calibration_set_id is None
        self._calibration_set_id = dqa.calibration_set_id
        if device is None:
            self._device = IQMDevice(server_device_metadata)
        else:
            # validate device compatibility with the calibration set
            if device.metadata != server_device_metadata:
                if self._use_default_calibration_set:
                    raise ValueError(
                        "The given 'device' is not compatible with the server default calibration set "
                        f"{self._calibration_set_id}."
                    )
                raise ValueError(f"The given 'device' is not compatible with calibration set {calibration_set_id}.")
            self._device = device
        self._run_sweep_timeout = run_sweep_timeout
        self._compiler_options = compiler_options if compiler_options is not None else CircuitCompilationOptions()

    @property
    def device(self) -> IQMDevice:
        """Returns the device used by the sampler."""
        return self._device

    def close_client(self):
        """Close IQMClient's session with the user authentication server. Discard the client."""
        if not self._client:
            return
        self._client.close_auth_session()
        self._client = None

    def run_sweep(  # type: ignore[override]
        self, program: cirq.Circuit, params: cirq.Sweepable, repetitions: int = 1
    ) -> list[IQMResult]:
        circuits, resolvers = self._resolve_parameters(program, params)
        results, metadata = self._send_circuits(
            circuits,
            repetitions=repetitions,
        )
        return [
            IQMResult(measurements=result, params=resolver, metadata=metadata)
            for result, resolver in zip(results, resolvers)
        ]

    def run_iqm_batch(self, programs: list[cirq.Circuit], repetitions: int = 1) -> list[IQMResult]:
        """Sends a batch of circuits to be executed.

        Running circuits in a batch is more efficient and hence completes quicker than running the circuits
        individually. Circuits run in a batch must all measure the same qubits.

        Args:
            programs: quantum circuits to execute
            repetitions: number of times the circuits are sampled

        Returns:
            results of the execution

        Raises:
            ValueError: circuits are not valid for execution
            CircuitExecutionError: something went wrong on the server
            APITimeoutError: server did not return the results in the allocated time
            RuntimeError: IQM client session has been closed

        """
        results, metadata = self._send_circuits(
            programs,
            repetitions=repetitions,
        )
        return [IQMResult(measurements=result, metadata=metadata) for result in results]

    def create_run_request(
        self, programs: cirq.Circuit | list[cirq.Circuit], *, params: cirq.Sweepable = None, repetitions: int = 1
    ) -> RunRequest:
        """Creates a run request without submitting it for execution.

        This takes the same parameters as :meth:`run` and :meth:`run_iqm_batch`, and can be used to check the
        run request that would be sent when calling those functions.

        Args:
            programs: quantum circuit(s) that would be executed when submitting the run request
            params: same as ``params`` for :meth:`run`, used only if ``programs`` is not a list
            repetitions: number of times the circuits are sampled

        Returns:
            the created run request

        """
        if isinstance(programs, cirq.Circuit):
            programs, _ = self._resolve_parameters(programs, params)

        serialized_circuits = [serialize_circuit(circuit) for circuit in programs]

        if not self._client:
            raise RuntimeError("Cannot submit circuits since session to IQM client has been closed.")

        different_calset_ids = set()
        for circuit in programs:
            if hasattr(circuit, "iqm_calibration_set_id"):
                if circuit.iqm_calibration_set_id != self._calibration_set_id:
                    different_calset_ids.add(circuit.iqm_calibration_set_id)
            else:
                different_calset_ids.add(None)
        if different_calset_ids:
            warnings.warn(
                f"Circuits have been decomposed/routed using calibration set(s) {different_calset_ids}, "
                f"different than the current calibration set {self._calibration_set_id} of this sampler. "
                f"Decompose/route the circuits using this sampler to ensure successful execution."
            )
        if self._use_default_calibration_set:
            default_calset_id = self._client.get_dynamic_quantum_architecture(None).calibration_set_id
            if self._calibration_set_id != default_calset_id:
                warnings.warn(
                    f"Server default calibration set has changed from {self._calibration_set_id} "
                    f"to {default_calset_id}. Use a new IQMSampler to decompose/route the circuits using "
                    "the new calibration set to ensure successful execution."
                )

        return self._client.create_run_request(
            serialized_circuits,
            calibration_set_id=self._calibration_set_id,
            shots=repetitions,
            options=self._compiler_options,
        )

    def _send_circuits(
        self,
        circuits: list[cirq.Circuit],
        repetitions: int = 1,
    ) -> tuple[list[dict[str, np.ndarray]], ResultMetadata]:
        """Sends a batch of circuits to be executed and retrieves the results.

        If a user interrupts the program while it is waiting for results, attempts to abort the submitted job.

        Args:
            circuits: quantum circuits to execute
            repetitions: number of shots to sample from each circuit

        Returns:
            circuit execution results, result metadata

        """
        run_request = self.create_run_request(circuits, repetitions=repetitions)
        job_id = self._client.submit_run_request(run_request)

        timeout_arg = [self._run_sweep_timeout] if self._run_sweep_timeout is not None else []

        try:
            results = self._client.wait_for_results(job_id, *timeout_arg)

        except KeyboardInterrupt:
            try:
                self._client.abort_job(job_id)
            except JobAbortionError as e:
                warnings.warn(f"Failed to abort job: {e}")
            finally:
                sys.exit()

        if results.measurements is None:
            raise RuntimeError("No measurements returned from IQM quantum computer.")

        return (
            [{k: np.array(v) for k, v in measurements.items()} for measurements in results.measurements],
            ResultMetadata(job_id, results.metadata.calibration_set_id, run_request),
        )

    @staticmethod
    def _resolve_parameters(
        program: cirq.Circuit, params: cirq.Sweepable
    ) -> tuple[list[cirq.Circuit], list[cirq.ParamResolver]]:
        resolvers = list(cirq.to_resolvers(params))
        circuits = [cirq.protocols.resolve_parameters(program, res) for res in resolvers] if resolvers else [program]
        return circuits, resolvers


@dataclass
class ResultMetadata:
    """Metadata for an IQM execution result.

    Attributes:
        job_id: ID of the computational job.
        calibration_set_id: Calibration set used for this :class:`.IQMResult`.
        request: Request made to run the job.

    """

    job_id: UUID
    calibration_set_id: UUID | None
    request: RunRequest


class IQMResult(cirq.ResultDict):
    """Stores the results of a quantum circuit execution on an IQM device.

    Args:
        params: Parameter resolver used for this circuit, if any.
        measurements: Maps measurement keys to measurement results, which are 2-D arrays of dtype bool.
            `shape == (repetitions, qubits)`.
        records: Maps measurement keys to measurement results, which are 3D arrays of dtype bool.
            `shape == (repetitions, instances, qubits)`.
        metadata: Metadata for the circuit execution results.

    """

    def __init__(
        self,
        *,
        params: cirq.ParamResolver | None = None,
        measurements: Mapping[str, np.ndarray] | None = None,
        records: Mapping[str, np.ndarray] | None = None,
        metadata: ResultMetadata,
    ) -> None:
        super().__init__(params=params, measurements=measurements, records=records)
        self.metadata = metadata



# Copyright 2024 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MOVE gate to be used on the IQM Star architecture."""

import numpy as np
from qiskit.circuit import Gate
import qiskit.quantum_info as qi

# MOVE gate has undefined phases, so we pick two arbitrary phases here
_phase_1 = np.exp(0.7j)
_phase_2 = 1.0  # np.exp(1.2j)
MOVE_GATE_UNITARY = [
    [1.0, 0.0, 0.0, 0.0],
    [0.0, 0.0, _phase_1, 0.0],
    [0.0, _phase_1.conj(), 0.0, 0.0],
    [0.0, 0.0, 0.0, _phase_2],
]
"""Unitary matrix for simulating the ideal MOVE gate.

This matrix is not a realistic description of MOVE, since it applies a zero phase on the moved
state, and acts as identity in the :math:`|11\rangle` subspace, thus being equal to the SWAP gate."""


class MoveGate(Gate):
    r"""The MOVE operation is a unitary population exchange operation between a qubit and a resonator.
    Its effect is only defined in the invariant subspace :math:`S = \text{span}\{|00\rangle, |01\rangle, |10\rangle\}`,
    where it swaps the populations of the states :math:`|01\rangle` and :math:`|10\rangle`.
    Its effect on the orthogonal subspace is undefined.

    MOVE has the following presentation in the subspace :math:`S`:

    .. math:: \text{MOVE}_S = |00\rangle \langle 00| + a |10\rangle \langle 01| + a^{-1} |01\rangle \langle 10|,

    where :math:`a` is an undefined complex phase that is canceled when the MOVE gate is applied a second time.

    To ensure that the state of the qubit and resonator has no overlap with :math:`|11\rangle`, it is
    recommended that no single qubit gates are applied to the qubit in between a
    pair of MOVE operations.

    .. note::
       The MOVE gate must always be be applied on the qubit and the resonator in the
       order ``[qubit, resonator]``, regardless of which component is currently holding the state.
    """

    def __init__(self, label=None):
        """Initializes the move gate"""
        super().__init__("move", 2, [], label=label)
        self.unitary = qi.Operator(MOVE_GATE_UNITARY)

    def _define(self):
        """This function is purposefully not defined so that that the Qiskit transpiler cannot accidentally
        decompose the MOVE gate into a sequence of other gates, instead it will throw an error.
        """
        return



# Copyright 2022 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Qiskit backend provider for IQM backends."""

from __future__ import annotations

from collections.abc import Callable
from importlib.metadata import PackageNotFoundError, version
from typing import Any
from uuid import UUID
import warnings

from iqm.iqm_client import Circuit, CircuitCompilationOptions, CircuitValidationError, IQMClient, RunRequest
from iqm.iqm_client.util import to_json_dict
from iqm.qiskit_iqm.fake_backends import IQMFakeAdonis
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase
from iqm.qiskit_iqm.iqm_job import IQMJob
from iqm.qiskit_iqm.qiskit_to_iqm import serialize_instructions
from qiskit import QuantumCircuit
from qiskit.providers import JobStatus, JobV1, Options

try:
    __version__ = version("qiskit-iqm")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


class IQMBackend(IQMBackendBase):
    """Backend for executing quantum circuits on IQM quantum computers.

    Args:
        client: client instance for communicating with an IQM server
        calibration_set_id: ID of the calibration set the backend will use.
            ``None`` means the IQM server will be queried for the current default
            calibration set.
        kwargs: optional arguments to be passed to the parent Backend initializer

    """

    def __init__(self, client: IQMClient, *, calibration_set_id: str | UUID | None = None, **kwargs):
        if calibration_set_id is not None and not isinstance(calibration_set_id, UUID):
            calibration_set_id = UUID(calibration_set_id)
        self._use_default_calibration_set = calibration_set_id is None
        architecture = client.get_dynamic_quantum_architecture(calibration_set_id)
        super().__init__(architecture, **kwargs)
        self.client: IQMClient = client
        self._max_circuits: int | None = None
        self.name = "IQM Backend"
        self._calibration_set_id = architecture.calibration_set_id

    @classmethod
    def _default_options(cls) -> Options:
        """Qiskit method for defining the default options for running the backend. We don't use them since they would
        not be documented here. Instead, we use the keyword arguments of the run method to pass options.
        """
        return Options()

    @property
    def max_circuits(self) -> int | None:
        """Maximum number of circuits that should be run in a single batch.

        Currently there is no hard limit on the number of circuits that can be executed in a single batch/job.
        However, some libraries like Qiskit Experiments use this property to split multi-circuit computational
        tasks into multiple baches/jobs.

        The default value is ``None``, meaning there is no limit. You can set it to a specific integer
        value to force these libraries to run at most that many circuits in a single batch.
        """
        return self._max_circuits

    @max_circuits.setter
    def max_circuits(self, value: int | None) -> None:
        self._max_circuits = value

    def run(
        self,
        run_input: QuantumCircuit | list[QuantumCircuit],
        **options,
    ) -> IQMJob:
        """Run a quantum circuit or a list of quantum circuits on the IQM quantum computer represented by this backend.

        Args:
            run_input: The circuits to run.
            options: Keyword arguments passed on to :meth:`create_run_request`, and documented there.

        Returns:
            Job object from which the results can be obtained once the execution has finished.

        """
        run_request = self.create_run_request(run_input, **options)
        job_id = self.client.submit_run_request(run_request)
        job = IQMJob(self, str(job_id), shots=run_request.shots)
        job.circuit_metadata = [c.metadata for c in run_request.circuits]
        return job

    def create_run_request(
        self,
        run_input: QuantumCircuit | list[QuantumCircuit],
        shots: int = 1024,
        circuit_compilation_options: CircuitCompilationOptions | None = None,
        circuit_callback: Callable[[list[QuantumCircuit]], Any] | None = None,
        qubit_mapping: dict[int, str] | None = None,
        **unknown_options,
    ) -> RunRequest:
        """Creates a run request without submitting it for execution.

        This can be used to check what would be submitted for execution by an equivalent call to :meth:`run`.

        Args:
            run_input: Same as in :meth:`run`.

        Args:
            shots: Number of repetitions of each circuit, for sampling.
            circuit_compilation_options:
                Compilation options for the circuits, passed on to :class:`~iqm.iqm_client.iqm_client.IQMClient`.
                If ``None``, the defaults of the :class:`~iqm.iqm_client.models.CircuitCompilationOptions`
                class are used.
            circuit_callback:
                Callback function that, if provided, will be called for the circuits before sending
                them to the device.  This may be useful in situations when you do not have explicit
                control over transpilation, but need some information on how it was done. This can
                happen, for example, when you use pre-implemented algorithms and experiments in
                Qiskit, where the implementation of the said algorithm or experiment takes care of
                delivering correctly transpiled circuits to the backend. This callback method gives
                you a chance to look into those transpiled circuits, and extract any info you need.
                As a side effect, you can also use this callback to modify the transpiled circuits
                in-place, just before execution; however, we do not recommend to use it for this
                purpose.
            qubit_mapping: Mapping from qubit indices in the circuit to qubit names on the device. If ``None``,
                :attr:`.IQMBackendBase.index_to_qubit_name` will be used.

        Returns:
            The created run request object

        """
        circuits = [run_input] if isinstance(run_input, QuantumCircuit) else run_input

        if len(circuits) == 0:
            raise ValueError("Empty list of circuits submitted for execution.")

        # Catch old iqm-client options
        if "max_circuit_duration_over_t2" in unknown_options or "heralding_mode" in unknown_options:
            warnings.warn(
                DeprecationWarning(
                    "max_circuit_duration_over_t2 and heralding_mode are deprecated, please use "
                    + "circuit_compilation_options instead."
                )
            )
        if circuit_compilation_options is None:
            cc_options_kwargs = {}
            if "max_circuit_duration_over_t2" in unknown_options:
                cc_options_kwargs["max_circuit_duration_over_t2"] = unknown_options.pop("max_circuit_duration_over_t2")
            if "heralding_mode" in unknown_options:
                cc_options_kwargs["heralding_mode"] = unknown_options.pop("heralding_mode")
            circuit_compilation_options = CircuitCompilationOptions(**cc_options_kwargs)

        if unknown_options:
            warnings.warn(f"Unknown backend option(s): {unknown_options}")

        if circuit_callback:
            circuit_callback(circuits)

        circuits_serialized: list[Circuit] = [self.serialize_circuit(circuit, qubit_mapping) for circuit in circuits]

        if self._use_default_calibration_set:
            default_calset_id = self.client.get_dynamic_quantum_architecture(None).calibration_set_id
            if self._calibration_set_id != default_calset_id:
                warnings.warn(
                    f"Server default calibration set has changed from {self._calibration_set_id} "
                    f"to {default_calset_id}. Create a new IQMBackend if you wish to transpile the "
                    "circuits using the new calibration set."
                )
        try:
            run_request = self.client.create_run_request(
                circuits_serialized,
                calibration_set_id=self._calibration_set_id,
                shots=shots,
                options=circuit_compilation_options,
            )
        except CircuitValidationError as e:
            raise CircuitValidationError(
                f"{e}\nMake sure the circuits have been transpiled using the same backend that you used to submit "
                f"the circuits."
            ) from e

        return run_request

    def retrieve_job(self, job_id: str) -> IQMJob:
        """Create and return an IQMJob instance associated with this backend with given job id.

        Args:
            job_id: ID of the job to retrieve.

        Returns:
            corresponding job

        """
        return IQMJob(self, job_id)

    def close_client(self) -> None:
        """Close IQMClient's session with the authentication server."""
        self.client.close_auth_session()

    def serialize_circuit(self, circuit: QuantumCircuit, qubit_mapping: dict[int, str] | None = None) -> Circuit:
        """Serialize a quantum circuit into the IQM data transfer format.

        Serializing is not strictly bound to the native gateset, i.e. some gates that are not explicitly mentioned in
        the native gateset of the backend can still be serialized. For example, the native single qubit gate for IQM
        backend is the 'r' gate, however 'x', 'rx', 'y' and 'ry' gates can also be serialized since they are just
        particular cases of the 'r' gate. If the circuit was transpiled against a backend using Qiskit's transpiler
        machinery, these gates are not supposed to be present. However, when constructing circuits manually and
        submitting directly to the backend, it is sometimes more explicit and understandable to use these concrete
        gates rather than 'r'. Serializing them explicitly makes it possible for the backend to accept such circuits.

        Qiskit uses one measurement instruction per qubit (i.e. there is no measurement grouping concept). While
        serializing we do not group any measurements together but rather associate a unique measurement key with each
        measurement instruction, so that the results can later be reconstructed correctly (see :class:`.MeasurementKey`
        documentation for more details).

        Args:
            circuit: quantum circuit to serialize
            qubit_mapping: Mapping from qubit indices in the circuit to qubit names on the device. If not provided,
                :attr:`.IQMBackendBase.index_to_qubit_name` will be used.

        Returns:
            data transfer object representing the circuit

        Raises:
            ValueError: circuit contains an unsupported instruction or is not transpiled in general

        """
        if qubit_mapping is None:
            qubit_mapping = self._idx_to_qb
        instructions = tuple(serialize_instructions(circuit, qubit_index_to_name=qubit_mapping))

        try:
            metadata = to_json_dict(circuit.metadata)
        except ValueError:
            warnings.warn(
                f"Metadata of circuit {circuit.name} was dropped because it could not be serialised to JSON.",
            )
            metadata = None

        return Circuit(name=circuit.name, instructions=instructions, metadata=metadata)


class IQMFacadeBackend(IQMBackend):
    """Facade backend for mimicking the execution of quantum circuits on IQM quantum computers. Allows to submit a
     circuit to the IQM server, and if the execution was successful, performs a simulation with a respective IQM noise
     model locally, then returns the simulated results.

    Args:
        client: client instance for communicating with an IQM server
        **kwargs: optional arguments to be passed to the parent Backend initializer

    """

    def __init__(self, client: IQMClient, **kwargs):
        self.fake_adonis = IQMFakeAdonis()
        target_architecture = client.get_static_quantum_architecture()

        if not self.fake_adonis.validate_compatible_architecture(target_architecture):
            raise ValueError("Quantum architecture of the remote quantum computer does not match Adonis.")

        super().__init__(client, **kwargs)
        self.name = "facade_adonis"

    def _validate_no_empty_cregs(self, circuit: QuantumCircuit) -> bool:
        """Returns True if given circuit has no empty (unused) classical registers, False otherwise."""
        cregs_utilization = dict.fromkeys(circuit.cregs, 0)
        used_cregs = [circuit.find_bit(i.clbits[0]).registers[0][0] for i in circuit.data if len(i.clbits) > 0]
        for creg in used_cregs:
            cregs_utilization[creg] += 1

        if 0 in cregs_utilization.values():
            return False
        return True

    def run(self, run_input: QuantumCircuit | list[QuantumCircuit], **options) -> JobV1:
        circuits = [run_input] if isinstance(run_input, QuantumCircuit) else run_input
        circuits_validated_cregs: list[bool] = [self._validate_no_empty_cregs(circuit) for circuit in circuits]
        if not all(circuits_validated_cregs):
            raise ValueError(
                "One or more circuits contain unused classical registers. This is not allowed for Facade simulation, "
                "see user guide."
            )

        iqm_backend_job = super().run(run_input, **options)
        iqm_backend_job.result()  # get and discard results
        if iqm_backend_job.status() == JobStatus.ERROR:
            raise RuntimeError("Remote execution did not succeed.")
        return self.fake_adonis.run(run_input, **options)


class IQMProvider:
    """Provider for IQM backends.

    IQMProvider connects to a quantum computer through an IQM server.
    If the server requires user authentication, you can provide it either using environment
    variables, or as keyword arguments to IQMProvider. The user authentication kwargs are passed
    through to :class:`~iqm.iqm_client.iqm_client.IQMClient` as is, and are documented there.

    Args:
        url: URL of the IQM server (e.g. https://cocos.resonance.meetiqm.com/garnet)

    """

    def __init__(self, url: str, **user_auth_args):  # contains keyword args auth_server_url, username, password
        self.url = url
        self.user_auth_args = user_auth_args

    def get_backend(
        self, name: str | None = None, calibration_set_id: UUID | None = None
    ) -> IQMBackend | IQMFacadeBackend:
        """An IQMBackend instance associated with this provider.

        Args:
            name: optional name of a custom facade backend
            calibration_set_id: ID of the calibration set used to create the transpilation target of the backend.
                If None, the server default calibration set will be used.

        """
        client = IQMClient(self.url, **self.user_auth_args)

        if name and name.startswith("facade_"):
            if name == "facade_adonis":
                return IQMFacadeBackend(client)

            warnings.warn(f"Unknown facade backend: {name}. A regular backend associated with {self.url} will be used.")

        return IQMBackend(client, calibration_set_id=calibration_set_id)



# Copyright 2020–2021 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Circuit optimization classes."""

import operator

import cirq
from cirq import circuits, ops


def simplify_circuit(
    circuit: cirq.Circuit,
    *,
    max_iterations: int = 20,
    drop_final_rz: bool = False,
) -> cirq.Circuit:
    """Simplifies and optimizes the given circuit.

    Currently it

    * merges any neighboring gates belonging to the same one-parameter family
    * merges all one-qubit rotations into phased X rotations followed by Z rotations
    * pushes the Z rotations towards the end of the circuit as far as possible
    * drops any empty Moments

    This sequence of optimization passes is repeated until the circuit hits a fixed point,
    or ``max_iterations`` is exceeded.

    Finally, it removes Z rotations that are immediately followed by a Z-basis measurement.

    Args:
        circuit: circuit to simplify
        max_iterations: maximum number of simplification rounds
        drop_final_rz: iff True, drop z rotations that have no successor operations

    Returns:
        simplified circuit

    """
    calset_id = getattr(circuit, "iqm_calibration_set_id", None)
    c = circuit.copy()

    # the optimizers cause the immediate decomposition of any gates they insert into the Circuit
    for _ in range(max_iterations):
        # It seems that Cirq optimizers have no way of communicating
        # if they actually made any changes to the Circuit.
        # See https://github.com/quantumlib/Cirq/issues/3761

        before = c.copy()

        # all mergeable 2-qubit gates are merged
        MergeOneParameterGroupGates().optimize_circuit(c)
        c = cirq.merge_single_qubit_gates_to_phased_x_and_z(c)
        # all z rotations are pushed past the first two-qubit gate following them
        c = cirq.eject_z(c, eject_parameterized=True)
        c = cirq.drop_empty_moments(c)

        if c == before:
            # the optimization hit a fixed point
            break

    DropRZBeforeMeasurement(drop_final=drop_final_rz).optimize_circuit(c)
    c = cirq.drop_empty_moments(c)
    c.iqm_calibration_set_id = calset_id  # type: ignore

    return c


class MergeOneParameterGroupGates(circuits.PointOptimizer):
    """Merges adjacent gates belonging to the same parametrized gate family.

    The merged gates have to act on the same sequence of qubits.
    This optimizer only works with gate families that are known to be one-parameter groups.

    For now, all the families are assumed to be periodic with a period of 4.
    """

    # TODO: ZZPowGate has a period of 2
    ONE_PARAMETER_FAMILIES = (ops.ISwapPowGate, ops.ZZPowGate)
    PERIOD = 4
    GATE_MERGING_TOLERANCE = 1e-10

    @classmethod
    def _normalize_par(cls, par):
        """Normalizes the given parameter value to (-period/2, period/2]."""
        shift = cls.PERIOD / 2
        return operator.mod(par - shift, -cls.PERIOD) + shift

    def optimization_at(
        self,
        circuit: cirq.Circuit,
        index: int,
        op: cirq.Operation,
    ) -> cirq.PointOptimizationSummary | None:
        """Describes how to change operations near the given location.

        Args:
            circuit: The circuit to improve.
            index: The index of the moment with the operation to focus on.
            op: The operation to focus improvements upon.

        Returns:
            A description of the optimization to perform, or else None if no
            change should be made.

        """
        if not isinstance(op.gate, self.ONE_PARAMETER_FAMILIES):
            return None

        def is_not_mergable(next_op: cirq.Operation) -> bool:
            """Predicate for finding gates that can be merged with op.

            A gate is mergable with op iff it (1) belongs to the same gate family,
            and (2) is acting on the same qubits.
            """
            if not isinstance(next_op.gate, type(op.gate)):
                return True
            if isinstance(op.gate, ops.gate_features.InterchangeableQubitsGate):
                # same qubits in any order
                return set(op.qubits) != set(next_op.qubits)
            # same qubits in the same order
            return op.qubits != next_op.qubits

        # start searching from op onwards
        start_frontier = {q: index for q in op.qubits}
        op_list = circuit.findall_operations_until_blocked(start_frontier, is_blocker=is_not_mergable)

        if len(op_list) == 1:
            return None  # just the one gate found, no changes
        indices, operations = zip(*op_list)

        # all the gates are in the same family so we may simply sum their parameters (mod periodicity)
        par = sum(o.gate.exponent for o in operations)
        # zero parameter (mod period) corresponds to identity
        # due to floating point errors we may be just a little below the period, which should also be
        # considered close to zero so let's shift away from the troublesome point before taking the modulo
        par = self._normalize_par(par)
        if abs(par) <= self.GATE_MERGING_TOLERANCE:
            rewritten: "cirq.OP_TREE" = []
        else:
            rewritten = op.gate.__class__(exponent=par).on(*op.qubits)

        return circuits.PointOptimizationSummary(
            clear_span=max(indices) + 1 - index, clear_qubits=op.qubits, new_operations=rewritten
        )


class DropRZBeforeMeasurement(circuits.PointOptimizer):
    """Drops z rotations that happen right before a z-basis measurement.

    These z rotations do not affect the result of the measurement, so we may ignore them.

    Args:
        drop_final: iff True, drop also any z rotation at the end of the circuit (since it's not
            followed by a measurement, it cannot affect them)

    """

    def __init__(self, drop_final: bool = False):
        super().__init__()
        self.drop_final = drop_final

    def optimization_at(
        self,
        circuit: cirq.Circuit,
        index: int,
        op: cirq.Operation,
    ) -> cirq.PointOptimizationSummary | None:
        """Describes how to change operations near the given location.

        Args:
            circuit: The circuit to improve.
            index: The index of the moment with the operation to focus on.
            op: The operation to focus improvements upon.

        Returns:
            A description of the optimization to perform, or else None if no
            change should be made.

        """

        def find_removable_rz() -> list[int]:
            """Finds z rotations that can be removed.

            A z rotation is removable iff it is followed by a z-basis measurement.

            Returns:
                moment indices of the z rotations to be removed

            """
            # op is a ZPowGate
            remove_indices = []
            for idx, moment in enumerate(circuit[index:], start=index):
                for x in moment.operations:
                    if op.qubits[0] in x.qubits:
                        # x acts on the same qubit as op
                        if isinstance(x.gate, cirq.ZPowGate):
                            # add idx to the list, keep looking for more
                            remove_indices.append(idx)
                            break  # to next moment
                        if isinstance(x.gate, cirq.MeasurementGate):
                            # follows the ZPowGates, remove the accumulated indices
                            return remove_indices
                        return []  # other operations: do not remove anything
            # circuit ends here
            if self.drop_final:
                return remove_indices
            return []

        if not isinstance(op.gate, cirq.ZPowGate):
            return None  # shortcut
        indices = find_removable_rz()
        if not indices:
            return None
        return circuits.PointOptimizationSummary(
            clear_span=max(indices) + 1 - index, clear_qubits=op.qubits, new_operations=[]
        )



# Copyright 2022-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Qiskit backend for IQM quantum computers."""

from __future__ import annotations

from abc import ABC
import itertools
from typing import Final

from iqm.iqm_client import (
    DynamicQuantumArchitecture,
    GateImplementationInfo,
    GateInfo,
)
from iqm.qiskit_iqm.move_gate import MoveGate
from qiskit.circuit import Delay, Parameter, Reset
from qiskit.circuit.library import CZGate, IGate, Measure, RGate
from qiskit.providers import BackendV2
from qiskit.transpiler import Target

IQM_TO_QISKIT_GATE_NAME: Final[dict[str, str]] = {"prx": "r", "cz": "cz"}

Locus = tuple[str, ...]
LocusIdx = tuple[int, ...]


class IQMBackendBase(BackendV2, ABC):
    """Abstract base class for various IQM-specific backends.

    Args:
        architecture: Description of the quantum architecture associated with the backend instance.

    """

    def __init__(
        self,
        architecture: DynamicQuantumArchitecture,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.architecture = architecture
        """Dynamic quantum architecture of the backend instance."""

        # Qiskit uses integer indices to refer to qubits, so we need to map component names to indices.
        # Because of the way the Target and the transpiler interact, the resonators need to have higher indices than
        # qubits, or else transpiling with optimization_level=0 will fail because of lacking resonator indices.
        qb_to_idx = {qb: idx for idx, qb in enumerate(architecture.qubits + architecture.computational_resonators)}

        self._target = IQMTarget(architecture, qb_to_idx, include_resonators=False)
        self._fake_target_with_moves = (
            IQMTarget(architecture, qb_to_idx, include_resonators=True) if "move" in architecture.gates else None
        )
        self._qb_to_idx = qb_to_idx
        self._idx_to_qb = {v: k for k, v in qb_to_idx.items()}
        self.name = "IQMBackend"
        self._coupling_map = self.target.build_coupling_map()

    @property
    def target(self) -> Target:
        return self._target

    @property
    def target_with_resonators(self) -> Target:
        """Return the target with MOVE gates and resonators included."""
        if self._fake_target_with_moves is None:
            return self.target
        return self._fake_target_with_moves

    @property
    def physical_qubits(self) -> list[str]:
        """Return the list of physical qubits in the backend."""
        return list(self._qb_to_idx)

    def has_resonators(self) -> bool:
        """True iff the backend QPU has computational resonators."""
        return bool(self.architecture.computational_resonators)

    def get_real_target(self) -> Target:
        """Return the real physical target of the backend without virtual CZ gates."""
        return IQMTarget(
            architecture=self.architecture,
            component_to_idx=self._qb_to_idx,
            include_resonators=True,
            include_fake_czs=False,
        )

    def qubit_name_to_index(self, name: str) -> int:
        """Given an IQM-style qubit name, return the corresponding index in the register.

        Args:
            name: IQM-style qubit name ('QB1', 'QB2', etc.)

        Returns:
            Index of the given qubit in the quantum register.

        Raises:
            ValueError: Qubit name cannot be found on the backend.

        """
        if name not in self._qb_to_idx:
            raise ValueError(f"Qubit '{name}' is not found on the backend.")
        return self._qb_to_idx[name]

    def index_to_qubit_name(self, index: int) -> str:
        """Given a quantum register index, return the corresponding IQM-style qubit name.

        Args:
            index: Qubit index in the quantum register.

        Returns:
            Corresponding IQM-style qubit name ('QB1', 'QB2', etc.).

        Raises:
            ValueError: Qubit index cannot be found on the backend.

        """
        if index not in self._idx_to_qb:
            raise ValueError(f"Qubit index {index} is not found on the backend.")
        return self._idx_to_qb[index]

    def get_scheduling_stage_plugin(self) -> str:
        """Return the plugin that should be used for scheduling the circuits on this backend."""
        return "iqm_default_scheduling"

    def restrict_to_qubits(
        self, qubits: list[int] | list[str], include_resonators: bool = False, include_fake_czs: bool = True
    ) -> IQMTarget:
        """Generated a restricted transpilation target from this backend that only contains the given qubits.

        Args:
            qubits: Qubits to restrict the target to. Can be either a list of qubit indices or qubit names.
            include_resonators: Whether to restrict `self.target` or `self.target_with_resonators`.
            include_fake_czs: Whether to include virtual CZs that are unsupported, but could be routed via MOVE.

        Returns:
            restricted target

        """
        qubits_str = [self._idx_to_qb[q] if isinstance(q, int) else str(q) for q in qubits]
        return _restrict_dqa_to_qubits(self.architecture, qubits_str, include_resonators, include_fake_czs)


def _restrict_dqa_to_qubits(
    architecture: DynamicQuantumArchitecture, qubits: list[str], include_resonators: bool, include_fake_czs: bool = True
) -> IQMTarget:
    """Generated a restricted transpilation target from this backend that only contains the given qubits.

    Args:
        architecture: The dynamic quantum architecture to restrict.
        qubits: Qubits to restrict the target to. Can be either a list of qubit indices or qubit names.
        include_resonators: Whether to include MOVE gates in the target.
        include_fake_czs: Whether to include virtual CZs that are not natively supported, but could be routed via MOVE.

    Returns:
        restricted target

    """
    new_gates = {}
    for gate_name, gate_info in architecture.gates.items():
        new_implementations = {}
        for implementation_name, implementation_info in gate_info.implementations.items():
            new_loci = tuple(locus for locus in implementation_info.loci if all(q in qubits for q in locus))
            if new_loci:
                new_implementations[implementation_name] = GateImplementationInfo(loci=new_loci)
        if new_implementations:
            new_gates[gate_name] = GateInfo(
                implementations=new_implementations,
                default_implementation=gate_info.default_implementation,
                override_default_implementation=gate_info.override_default_implementation,
            )
    new_arch = DynamicQuantumArchitecture(
        calibration_set_id=architecture.calibration_set_id,
        qubits=[q for q in qubits if q in architecture.qubits],
        computational_resonators=[q for q in qubits if q in architecture.computational_resonators],
        gates=new_gates,
    )
    return IQMTarget(new_arch, {name: idx for idx, name in enumerate(qubits)}, include_resonators, include_fake_czs)


class IQMTarget(Target):
    """Transpilation target for an IQM architecture.

    Contains the mapping of physical qubit name on the device to qubit index in the Target.

    Args:
        architecture: Quantum architecture that defines the target.
        component_to_idx: Mapping from QPU component names to integer indices used by Qiskit to refer to them.
        include_resonators: Whether to include MOVE gates in the target.
        include_fake_czs: Whether to include virtual CZs that are not natively supported, but could be routed via MOVE.

    """

    def __init__(
        self,
        architecture: DynamicQuantumArchitecture,
        component_to_idx: dict[str, int],
        include_resonators: bool,
        include_fake_czs: bool = True,
    ):
        super().__init__()
        # Using iqm as a prefix to avoid name clashes with the base class.
        self.iqm_dqa = architecture
        self.iqm_component_to_idx = component_to_idx
        self.iqm_idx_to_component = {v: k for k, v in component_to_idx.items()}
        self.iqm_includes_resonators = include_resonators
        self.iqm_includes_fake_czs = include_fake_czs
        self._add_connections_from_DQA()

    def _add_connections_from_DQA(self):
        """Initializes the Target, making it represent the dynamic quantum architecture :attr:`iqm_dqa`."""
        # mapping from op name to all its allowed loci
        architecture = self.iqm_dqa
        component_to_idx = self.iqm_component_to_idx
        op_loci = {gate_name: gate_info.loci for gate_name, gate_info in architecture.gates.items()}

        def locus_to_idx(locus: Locus) -> LocusIdx:
            """Map the given locus to use component indices instead of component names."""
            return tuple(component_to_idx[component] for component in locus)

        def create_properties(name: str, *, symmetrize: bool = False) -> dict[tuple[int, ...], None]:
            """Creates the Qiskit instruction properties dictionary for the given IQM native operation.

            Currently we do not provide any actual properties for the operation, hence the all the
            allowed loci map to None.
            """
            if self.iqm_includes_resonators:
                loci = op_loci[name]
            else:
                # Remove the loci that correspond to resonators.
                loci = [
                    locus for locus in op_loci[name] if all(component in self.iqm_dqa.qubits for component in locus)
                ]
            if symmetrize:
                # symmetrize the loci
                loci = tuple(permuted_locus for locus in loci for permuted_locus in itertools.permutations(locus))
            return {locus_to_idx(locus): None for locus in loci}

        # like barrier, delay is always available for all single-qubit loci
        self.add_instruction(Delay(0), {locus_to_idx((q,)): None for q in architecture.qubits})

        if "measure" in op_loci:
            self.add_instruction(Measure(), create_properties("measure"))

        # identity gate does nothing and is removed in serialization, so we may as well allow it everywhere
        # Except if it is defined for the resonator, the graph is disconnected and the transpiler will fail.
        if self.iqm_includes_resonators:
            self.add_instruction(
                IGate(),
                {locus_to_idx((component,)): None for component in architecture.components},
            )
        else:
            self.add_instruction(
                IGate(),
                {locus_to_idx((component,)): None for component in architecture.qubits},
            )

        if "prx" in op_loci:
            self.add_instruction(
                RGate(Parameter("theta"), Parameter("phi")),
                create_properties("prx"),
            )

        # HACK reset gate shares cc_prx loci for now
        if "cc_prx" in op_loci:
            self.add_instruction(Reset(), create_properties("cc_prx"))

        if self.iqm_includes_resonators and "move" in op_loci:
            self.add_instruction(MoveGate(), create_properties("move"))

        if "cz" in op_loci:
            if self.iqm_includes_fake_czs and "move" in op_loci:
                # CZ and MOVE: star
                cz_connections: dict[LocusIdx, None] = {}
                cz_loci = op_loci["cz"]
                for c1, c2 in cz_loci:
                    if self.iqm_includes_resonators or all(component in self.iqm_dqa.qubits for component in (c1, c2)):
                        idx_locus = locus_to_idx((c1, c2))
                        cz_connections[idx_locus] = None

                for c1, res in op_loci["move"]:
                    for c2 in architecture.qubits:
                        if c2 not in [c1, res]:
                            # loop over c2 that is not c1
                            if (c2, res) in cz_loci or (res, c2) in cz_loci:
                                # This is a fake CZ and can be bidirectional.
                                # cz routable via res between qubits, put into fake_cz_conn both ways
                                idx_locus = locus_to_idx((c1, c2))
                                cz_connections[idx_locus] = None
                                cz_connections[idx_locus[::-1]] = None
                self.add_instruction(CZGate(), cz_connections)
            else:
                # CZ but no MOVE: crystal
                self.add_instruction(CZGate(), create_properties("cz"))

    @property
    def physical_qubits(self) -> list[str]:
        """Return the ordered list of physical qubits in the backend."""
        # Overrides the property from the superclass to contain the correct information.
        return [self.iqm_idx_to_component[i] for i in range(self.num_qubits)]

    def restrict_to_qubits(self, qubits: list[int] | list[str]) -> IQMTarget:
        """Generated a restricted transpilation target from this Target that only contains the given qubits.

        Args:
            qubits: Qubits to restrict the target to. Can be either a list of qubit indices or qubit names.

        Returns:
            restricted target

        """
        qubits_str = [self.iqm_idx_to_component[q] if isinstance(q, int) else str(q) for q in qubits]
        return _restrict_dqa_to_qubits(
            self.iqm_dqa, qubits_str, self.iqm_includes_resonators, self.iqm_includes_fake_czs
        )



# Copyright 2023 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file is an example of using Qiskit on IQM to run a simple but non-trivial quantum circuit on an IQM quantum
computer. See the Qiskit on IQM user guide for instructions:
https://docs.meetiqm.com/iqm-client/user_guide_qiskit.html
"""

import argparse

from iqm.qiskit_iqm.iqm_provider import IQMProvider
from qiskit import QuantumCircuit, transpile


def transpile_example(server_url: str) -> tuple[QuantumCircuit, dict[str, int]]:
    """Run a GHZ circuit transpiled using the Qiskit transpile function.

    Args:
        server_url: URL of the IQM server used for execution

    Returns:
        transpiled circuit, a mapping of bitstrings representing qubit measurement results to counts for each result

    """
    backend = IQMProvider(server_url).get_backend()

    num_qubits = min(backend.num_qubits, 5)  # use at most 5 qubits
    circuit = QuantumCircuit(num_qubits)
    circuit.h(0)
    for i in range(1, num_qubits):
        circuit.cx(0, i)
    circuit.measure_all()

    transpiled_circuit = transpile(circuit, backend)
    counts = backend.run(transpiled_circuit, shots=1000).result().get_counts()

    return transpiled_circuit, counts


if __name__ == "__main__":
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "--url",
        help="URL of the IQM service",
        default="https://cocos.resonance.meetiqm.com/deneb",
    )
    circuit_transpiled, results = transpile_example(argparser.parse_args().url)
    print(circuit_transpiled)
    print(results)



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementations for IQM specific quantum gates"""

from cirq import CircuitDiagramInfo, CircuitDiagramInfoArgs, EigenGate
import numpy as np


class IQMMoveGate(EigenGate):
    r"""The MOVE operation is a unitary population exchange operation between a qubit and a resonator.
    Its effect is only defined in the invariant subspace :math:`S = \text{span}\{|00\rangle, |01\rangle, |10\rangle\}`,
    where it swaps the populations of the states :math:`|01\rangle` and :math:`|10\rangle`.
    Its effect on the orthogonal subspace is undefined.

    MOVE has the following presentation in the subspace :math:`S`:

    .. math:: \text{MOVE}_S = |00\rangle \langle 00| + a |10\rangle \langle 01| + a^{-1} |01\rangle \langle 10|,

    where :math:`a` is an undefined complex phase that is canceled when the MOVE gate is applied a second time.

    To ensure that the state of the qubit and resonator has no overlap with :math:`|11\rangle`, it is
    recommended that no single qubit gates are applied to the qubit in between a
    pair of MOVE operations.

    Note: At this point the locus for the move gate must be defined in the order: ``[qubit, resonator]``.
    Additionally, the matrix representation of the gate set to be a SWAP gate, even though this is not what physically
    happens.
    """

    def _num_qubits_(self) -> int:
        return 2

    def _eigen_components(self) -> list[tuple[float, np.ndarray]]:
        return [
            (0, np.array([[1, 0, 0, 0], [0, 0.5, 0.5, 0], [0, 0.5, 0.5, 0], [0, 0, 0, 1]])),
            (1, np.array([[0, 0, 0, 0], [0, 0.5, -0.5, 0], [0, -0.5, 0.5, 0], [0, 0, 0, 0]])),
        ]

    def _circuit_diagram_info_(self, args: CircuitDiagramInfoArgs) -> CircuitDiagramInfo:
        return CircuitDiagramInfo(wire_symbols=("MOVE(QB)", "MOVE(Res)"), exponent=self._diagram_exponent(args))

    def __str__(self) -> str:
        return "MOVE"



# Copyright 2021-2023 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Token manager for authentication and authorization to IQM's quantum computers. Part of IQM Client CLI."""

from datetime import datetime
import json
import os
from pathlib import Path
import time

from iqm.iqm_client.cli.auth import AUTH_REQUESTS_TIMEOUT, ClientAuthenticationError, refresh_request
from iqm.iqm_client.cli.models import ConfigFile
from psutil import pid_exists
from requests.exceptions import ConnectionError, Timeout


def daemonize_token_manager(cycle: int, config: ConfigFile, logfile: str | None = None) -> None:
    """Start a daemon process.

    Args:
        cycle: refresh cycle in seconds
        config: IQM Client CLI configuration
        logfile: path to file for writing errors

    """
    import daemon

    logfile = (
        logfile
        if logfile is not None
        else f"{os.environ.get('XDG_STATE_HOME', f'{Path.home()}/.local/state')}/iqm-client-cli/token_manager.log"
    )
    os.makedirs(os.path.dirname(logfile), exist_ok=True)
    with open(logfile, "w", encoding="UTF-8") as output:
        with daemon.DaemonContext(stdout=output, stderr=output):
            start_token_manager(cycle, config)


def start_token_manager(cycle: int, config: ConfigFile, single_run: bool = False) -> None:
    """Refresh tokens periodically.

    For each refresh cycle new tokens are requested from auth server.
    - If refresh is successful next refresh is attempted in the next cycle.
    - If auth server does not respond refresh is attempted repeatedly until it succeeds or
      the existing refresh token expires.
    - If auth server responds but returns an error code or invalid tokens token manager is stopped.

    Args:
        cycle: refresh cycle in seconds
        config: IQM Client CLI configuration
        single_run: if True, refresh tokens only once and exit; otherwise repeat refreshing indefinitely

    """
    while True:
        tokens_file = str(config.tokens_file)
        tokens = read_tokens(tokens_file)

        new_tokens, status, sleep_time = refresh_tokens(config, tokens, cycle)
        if new_tokens is None:
            break

        write_tokens(tokens_file, str(config.auth_server_url), status, **new_tokens)

        if single_run:
            break

        time.sleep(sleep_time)

    print(f"{datetime.now().strftime('%m/%d/%Y %H:%M:%S')}: Token manager stopped")


def read_tokens(path_to_tokens_file: str) -> dict:
    """Read current tokens from the tokens file.

    Args:
        path_to_tokens_file: path to the tokens file

    Returns:
        dict containing the tokens

    """
    with open(path_to_tokens_file, "r", encoding="utf-8") as file:
        tokens = json.load(file)
    return tokens


def refresh_tokens(config: ConfigFile, current_tokens: dict, cycle: int) -> tuple[dict | None, bool, int]:
    """Request new tokens from auth server.

    Args:
        config: IQM Client CLI configuration
        current_tokens: dict containing the current tokens from the tokens file
        cycle: refresh cycle length in seconds

    Returns:
        Tuple[Optional[dict], bool, int] = (tokens, status, sleep_time)
        tokens: dict containing new tokens or current tokens if auth server could not be connected or
                None if auth server refused to provide new tokens.
        status: bool, True if tokens were refreshed successfully, False otherwise
        sleep_time: time to sleep before next refresh attempt

    """
    access_token = current_tokens.get("access_token", "")
    refresh_token = current_tokens.get("refresh_token", "")
    try:
        tokens = refresh_request(str(config.auth_server_url), config.realm, config.client_id, refresh_token)
        status = True
        sleep_time = cycle
        log_timestamp = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
        print(f"{log_timestamp}: Tokens refreshed successfully.")
    except (Timeout, ConnectionError) as ex:
        # No connection to auth server or auth server did not respond, keep current tokens
        tokens = {"access_token": access_token, "refresh_token": refresh_token}
        status = False
        if isinstance(ex, ConnectionError):
            sleep_time = AUTH_REQUESTS_TIMEOUT
        else:
            sleep_time = 1
        log_timestamp = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
        print(f"{log_timestamp}: No response from auth server: {ex}")
    except ClientAuthenticationError as ex:
        # Auth server responded but no valid tokens were received
        tokens = None
        status = False
        sleep_time = cycle
        log_timestamp = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
        print(f"{log_timestamp}: Failed to authenticate with auth server: {ex}")

    return tokens, status, sleep_time


def write_tokens(
    path_to_tokens_file: str,
    auth_server_url: str,
    status: bool,
    *,
    access_token: str = "",
    refresh_token: str = "",
) -> None:
    """Write new tokens into the tokens file.

    Args:
        path_to_tokens_file: path to the tokens file
        auth_server_url: base URL of the auth server
        status: refresh status, True when successful, False otherwise
        access_token: new access token
        refresh_token: new refresh token

    """
    tokens_json = json.dumps(
        {
            "pid": os.getpid(),
            "timestamp": datetime.now().isoformat(),
            "refresh_status": "SUCCESS" if status else "FAILED",
            "access_token": access_token,
            "refresh_token": refresh_token,
            "auth_server_url": auth_server_url,
        }
    )

    try:
        Path(path_to_tokens_file).parent.mkdir(parents=True, exist_ok=True)
        with open(Path(path_to_tokens_file), "w", encoding="UTF-8") as file:
            file.write(tokens_json)
    except OSError as error:
        print("Error writing tokens file", error)


def check_token_manager(tokens_file: str) -> int | None:
    """Check whether a token manager related to the given tokens_file is running.

    Args:
        tokens_file: Path to a tokens JSON file.

    Returns:
        Optional[int]: PID of the process if process is running, None otherwise.

    """
    with open(tokens_file, "r", encoding="utf-8") as file:
        tokens_data = json.load(file)
    pid = tokens_data["pid"] if "pid" in tokens_data else None

    if pid and pid_exists(pid):
        return pid
    return None



# Copyright 2021-2023 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pydantic models for IQM Client CLI files."""

from datetime import datetime
from pathlib import Path

from pydantic import AnyUrl, BaseModel


class ConfigFile(BaseModel):
    """Model of configuration file, used for validating JSON."""

    auth_server_url: AnyUrl
    realm: str
    client_id: str
    username: str | None = None
    tokens_file: Path


class TokensFile(BaseModel):
    """Model of tokens file, used for validating JSON."""

    pid: int | None = None
    timestamp: datetime
    access_token: str
    refresh_token: str
    refresh_status: str | None = None
    auth_server_url: AnyUrl



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for IQM specific transpilation needs."""

from __future__ import annotations

from typing import TYPE_CHECKING

from cirq import Circuit
from iqm.cirq_iqm.serialize import deserialize_circuit, serialize_circuit
from iqm.iqm_client import ExistingMoveHandlingOptions, transpile_insert_moves

if TYPE_CHECKING:
    from iqm.cirq_iqm.devices import IQMDevice


def transpile_insert_moves_into_circuit(
    cirq_circuit: Circuit,
    device: IQMDevice,
    existing_moves: ExistingMoveHandlingOptions = ExistingMoveHandlingOptions.KEEP,
    qubit_mapping: dict[str, str] | None = None,
) -> Circuit:
    """Transpile the circuit to insert MOVE gates where needed.

    Args:
        cirq_circuit: Circuit to transpile.
        device: Device to transpile for.
        existing_moves: How to handle existing MOVE gates, obtained from the IQM client library.
        qubit_mapping: Mapping from qubit names in the circuit to the device.

    Returns:
        Transpiled circuit.

    """
    if device.metadata.architecture is None:
        raise ValueError("MOVE transpilation only supported for devices created from a dynamic quantum architecture.")
    iqm_client_circuit = serialize_circuit(cirq_circuit)
    new_iqm_client_circuit = transpile_insert_moves(
        iqm_client_circuit,
        device.metadata.architecture,
        existing_moves=existing_moves,
        qubit_mapping=qubit_mapping,
    )
    new_cirq_circuit = deserialize_circuit(new_iqm_client_circuit)
    return new_cirq_circuit



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""IQM's Apollo quantum architecture."""

from __future__ import annotations

from .iqm_device import IQMDevice, IQMDeviceMetadata


class Apollo(IQMDevice):
    r"""IQM's twenty-qubit transmon device.

    The qubits are connected thus::

            QB20  QB17
            /  \  /  \
         QB19  QB16  QB12
         /  \  /  \  /  \
      QB18  QB15  QB11  QB7
         \  /  \  /  \  /
         QB14  QB10  QB6
         /  \  /  \  /
      QB13  QB9   QB5
         \  /  \  /  \
         QB8   QB4   QB2
            \  /  \  /
            QB3   QB1

    where the lines denote which qubit pairs can be subject to two-qubit gates.

    Apollo has native PhasedXPowGate, XPowGate, and YPowGate gates. The two-qubit gate CZ is
    native, as well. The qubits can be measured simultaneously or separately any number of times.
    """

    def __init__(self):
        qubit_count = 20
        connectivity = (
            {1, 2},
            {1, 4},
            {2, 5},
            {3, 4},
            {3, 8},
            {4, 5},
            {4, 9},
            {5, 6},
            {5, 10},
            {6, 7},
            {6, 11},
            {7, 12},
            {8, 9},
            {8, 13},
            {9, 10},
            {9, 14},
            {10, 11},
            {10, 15},
            {11, 12},
            {11, 16},
            {12, 17},
            {13, 14},
            {14, 15},
            {14, 18},
            {15, 16},
            {15, 19},
            {16, 17},
            {16, 20},
            {18, 19},
            {19, 20},
        )
        super().__init__(IQMDeviceMetadata.from_qubit_indices(qubit_count, connectivity))



# Copyright 2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains user authentication related classes and functions required by IQMClient."""

from abc import ABC, abstractmethod
from base64 import b64decode
import json
import os
import time
from typing import Any

from iqm.iqm_client.errors import ClientAuthenticationError, ClientConfigurationError
import requests

AUTH_CLIENT_ID = "iqm_client"
AUTH_REALM = "cortex"
AUTH_REQUESTS_TIMEOUT = float(os.environ.get("IQM_CLIENT_REQUESTS_TIMEOUT", 60.0))
REFRESH_MARGIN_SECONDS = 60


class TokenManager:
    """TokenManager manages the access token required for user authentication.

    Args:
        token: Long-lived IQM token in plain text format
        tokens_file: Path to a tokens file used for authentication
        auth_server_url: Base URL of the authentication server
        username: Username to log in to authentication server
        password: Password to log in to authentication server

    The parameters can also be read from the environment variables IQM_TOKEN, IQM_TOKENS_FILE,
    IQM_AUTH_SERVER, IQM_AUTH_USERNAME, IQM_AUTH_PASSWORD. Environment variables can not be
    mixed with initialisation arguments. All parameters must come from the same source.

    """

    @staticmethod
    def time_left_seconds(token: Any) -> int:
        """Check how much time is left until the token expires.

        Returns:
            Time left on token in seconds.

        """
        if not token or not isinstance(token, str):
            return 0
        parts = token.split(".", 2)
        if len(parts) != 3:
            return 0
        # Add padding to adjust body length to a multiple of 4 chars as required by base64 decoding
        try:
            body = parts[1] + ("=" * (-len(parts[1]) % 4))
            exp_time = int(json.loads(b64decode(body)).get("exp", "0"))
            return max(0, exp_time - int(time.time()))
        except (UnicodeDecodeError, json.decoder.JSONDecodeError, ValueError, TypeError):
            return 0

    def __init__(
        self,
        token: str | None = None,
        tokens_file: str | None = None,
        auth_server_url: str | None = None,
        username: str | None = None,
        password: str | None = None,
    ):
        def _format_names(variable_names: list[str]) -> str:
            """Format a list of variable names"""
            return ", ".join(f'"{name}"' for name in variable_names)

        auth_parameters: dict[str, str] = {}

        init_parameters = {
            "token": token,
            "tokens_file": tokens_file,
            "auth_server_url": auth_server_url,
            "username": username,
            "password": password,
        }
        init_params_given = [key for key, value in init_parameters.items() if value]

        env_variables = {
            "token": "IQM_TOKEN",
            "tokens_file": "IQM_TOKENS_FILE",
            "auth_server_url": "IQM_AUTH_SERVER",
            "username": "IQM_AUTH_USERNAME",
            "password": "IQM_AUTH_PASSWORD",
        }
        env_vars_given = [name for name in env_variables.values() if os.environ.get(name)]

        if init_params_given and env_vars_given:
            raise ClientConfigurationError(
                "Authentication parameters given both as initialisation args and as environment variables: "
                + f"initialisation args {_format_names(init_params_given)}, "
                + f"environment variables {_format_names(env_vars_given)}."
                + " Parameter sources must not be mixed."
            )

        if env_vars_given:
            auth_parameters = {key: value for key, name in env_variables.items() if (value := os.environ.get(name))}
        else:
            auth_parameters = {key: str(value) for key, value in init_parameters.items() if value}

        self._token_provider: TokenProviderInterface | None = None
        self._access_token: str | None = None

        if not auth_parameters:
            self._token_provider = None
        elif set(auth_parameters) == {"token"}:
            # This is not necessarily a JWT token
            self._token_provider = ExternalToken(auth_parameters["token"])
        elif set(auth_parameters) == {"tokens_file"}:
            self._token_provider = TokensFileReader(auth_parameters["tokens_file"])
        elif set(auth_parameters) == {"auth_server_url", "username", "password"}:
            self._token_provider = TokenClient(
                auth_parameters["auth_server_url"],
                AUTH_REALM,
                auth_parameters["username"],
                auth_parameters["password"],
            )
        else:
            raise ClientConfigurationError(
                f"Invalid combination of authentication parameters specified: {list(auth_parameters)}",
            )

    def get_bearer_token(self, retries: int = 1) -> str | None:
        """Returns a valid bearer token, or None if no user authentication has been configured.

        Raises:
            ClientAuthenticationError: getting the token failed

        """
        if self._token_provider is None:
            return None  # Authentication is not used

        # Use the existing access token if it is still valid
        if TokenManager.time_left_seconds(self._access_token) > REFRESH_MARGIN_SECONDS:
            return f"Bearer {self._access_token}"

        # Otherwise, get a new access token from token provider
        try:
            self._access_token = self._token_provider.get_token()
            return f"Bearer {self._access_token}"
        except ClientAuthenticationError:
            if retries < 1:
                raise

        # Try again
        return self.get_bearer_token(retries - 1)

    def close(self) -> bool:
        """Close the configured token provider.

        Returns:
            True if closing was successful

        Raises:
            ClientAuthenticationError: closing failed

        """
        if self._token_provider is None:
            return False

        self._token_provider.close()
        self._token_provider = None
        return True


class TokenProviderInterface(ABC):
    """Interface to token provider"""

    @abstractmethod
    def get_token(self) -> str:
        """Returns a valid access token.

        Raises:
            ClientAuthenticationError: acquiring the token failed

        """

    @abstractmethod
    def close(self) -> None:
        """Closes authentication session.

        Raises:
            ClientAuthenticationError: closing the session failed

        """


class ExternalToken(TokenProviderInterface):
    """Holds an external token"""

    def __init__(self, token: str):
        self._token: str | None = token

    def get_token(self) -> str:
        if self._token is None:
            raise ClientAuthenticationError("No external token available")
        return self._token

    def close(self) -> None:
        self._token = None
        raise ClientAuthenticationError("Can not close externally managed auth session")


class TokensFileReader(TokenProviderInterface):
    """Reads token from a file"""

    def __init__(self, tokens_file: str):
        self._path: str | None = tokens_file

    def get_token(self) -> str:
        try:
            if self._path is None:
                raise ClientAuthenticationError("No tokens file available")
            with open(self._path, encoding="utf-8") as file:
                raw_data = file.read()
            json_data = json.loads(raw_data)
            token = json_data.get("access_token")
            if TokenManager.time_left_seconds(token) <= 0:
                raise ClientAuthenticationError("Access token in file has expired or is not valid")
        except (FileNotFoundError, IsADirectoryError, json.decoder.JSONDecodeError) as e:
            raise ClientAuthenticationError(rf"Failed to read access token from file '{self._path}': {e}") from e
        return token

    def close(self) -> None:
        self._path = None
        raise ClientAuthenticationError("Can not close externally managed auth session")


class TokenClient(TokenProviderInterface):
    """Requests new token from an authentication server"""

    PASSWORD_GRANT_TYPE = "password"
    REFRESH_TOKEN_GRANT_TYPE = "refresh_token"

    def __init__(self, auth_server_url: str, realm: str, username: str, password: str):
        """Initialize the client"""
        self._token_url = f"{auth_server_url}/realms/{realm}/protocol/openid-connect/token"
        self._logout_url = f"{auth_server_url}/realms/{realm}/protocol/openid-connect/logout"
        self._username = username
        self._password = password
        self._refresh_token: str | None = None

    def _get_access_token_from_server(self, grant_type: str) -> str | None:
        """Get new access token from the server and update refresh token."""
        if grant_type == TokenClient.REFRESH_TOKEN_GRANT_TYPE:
            # Update tokens using existing refresh_token
            data = {
                "client_id": AUTH_CLIENT_ID,
                "grant_type": TokenClient.REFRESH_TOKEN_GRANT_TYPE,
                "refresh_token": str(self._refresh_token),
            }
        else:
            # There is no valid refresh token or refresh token has expired, start a new session
            data = {
                "client_id": AUTH_CLIENT_ID,
                "grant_type": TokenClient.PASSWORD_GRANT_TYPE,
                "username": self._username,
                "password": self._password,
            }

        # Request new tokens from the server
        access_token: str | None = None
        result = requests.post(self._token_url, data=data, timeout=AUTH_REQUESTS_TIMEOUT)
        if result.status_code == 200:
            tokens = result.json()
            self._refresh_token = tokens.get("refresh_token")
            if TokenManager.time_left_seconds(self._refresh_token) <= 0:
                self._refresh_token = None
            access_token = tokens.get("access_token")
            if TokenManager.time_left_seconds(access_token) <= 0:
                access_token = None
        return access_token

    def get_token(self) -> str:
        """Get new access token and refresh token from the server"""
        if not self._token_url:
            raise ClientConfigurationError("No auth server configured")

        access_token: str | None = None
        if TokenManager.time_left_seconds(self._refresh_token) > REFRESH_MARGIN_SECONDS:
            # There is a valid refresh token, try to update tokens using it
            access_token = self._get_access_token_from_server(TokenClient.REFRESH_TOKEN_GRANT_TYPE)
        if access_token is None:
            # Failed to get valid access token using refresh token, start a new session
            access_token = self._get_access_token_from_server(TokenClient.PASSWORD_GRANT_TYPE)
        if access_token is None:
            # Failed to get valid access token using username and password, raise an error
            raise ClientAuthenticationError("Getting access token from auth server failed")
        return str(access_token)  # acces token can not be None here

    def close(self) -> None:
        """Close authentication session"""
        if not self._refresh_token:
            raise ClientAuthenticationError("No auth session active")

        data = {"client_id": AUTH_CLIENT_ID, "refresh_token": self._refresh_token}
        self._refresh_token = None

        result = requests.post(self._logout_url, data=data, timeout=AUTH_REQUESTS_TIMEOUT)
        if result.status_code not in [200, 204]:
            raise ClientAuthenticationError(f"Logout failed, {result.text}")



# Copyright 2022-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Conversion tools from Qiskit to IQM representation."""

from __future__ import annotations

from collections.abc import Collection
from dataclasses import dataclass
import re

from iqm.iqm_client import Instruction
from iqm.qiskit_iqm.move_gate import MoveGate
import numpy as np
from qiskit import QuantumCircuit as QiskitQuantumCircuit
from qiskit.circuit import ClassicalRegister, Clbit, QuantumRegister
from qiskit.transpiler.layout import Layout


class InstructionNotSupportedError(RuntimeError):
    """Raised when a given instruction is not supported by the IQM server."""


@dataclass(frozen=True)
class MeasurementKey:
    """Unique key associated with a measurement instruction.

    Qiskit stores the results of quantum measurements in classical registers consisting of bits.
    The circuit execution results are presented as bitstrings of a certain structure so that the classical
    register and the index within that register for each bit is implied from its position in the bitstring.

    For example, if you have two classical registers in the circuit with lengths 3 and 2, then the
    measurement results will look like '01 101' if the classical register of length 3 was added to
    the circuit first, and '101 01' otherwise. If a bit in a classical register is not used in any
    measurement operation it will still show up in the results with the default value of '0'.

    To be able to handle measurement results in a Qiskit-friendly way, we need to keep around some
    information about how the circuit was constructed. This can, for example, be achieved by keeping
    around the original Qiskit quantum circuit and using it when constructing the results in
    :class:`.IQMJob`. This should be done so that the circuit is saved on the server side and not in
    ``IQMJob``, since otherwise users will not be able to retrieve results from a detached Python
    environment solely based on the job id. Another option is to use measurement key strings to
    store the required info. Qiskit does not use measurement keys, so we are free to use them
    internally in the communication with the IQM server, and can encode the necessary information in
    them.

    This class encapsulates the necessary info, and provides methods to transform between this
    representation and the measurement key string representation.

    Args:
        creg_name: name of the classical register
        creg_len: number of bits in the classical register
        creg_idx: Index of the classical register in the circuit. Determines the order in which this register was added
            to the circuit relative to the others.
        clbit_idx: index of the classical bit within the classical register

    """

    creg_name: str
    creg_len: int
    creg_idx: int
    clbit_idx: int

    def __str__(self):
        return f"{self.creg_name}_{self.creg_len}_{self.creg_idx}_{self.clbit_idx}"

    @classmethod
    def from_string(cls, string: str) -> MeasurementKey:
        """Create a MeasurementKey from its string representation."""
        match = re.match(r"^(.*)_(\d+)_(\d+)_(\d+)$", string)
        if match is None:
            raise ValueError("Invalid measurement key string representation.")
        return cls(match.group(1), int(match.group(2)), int(match.group(3)), int(match.group(4)))

    @classmethod
    def from_clbit(cls, clbit: Clbit, circuit: QiskitQuantumCircuit) -> MeasurementKey:
        """Create a MeasurementKey for a classical bit in a quantum circuit."""
        bitloc = circuit.find_bit(clbit)
        creg = bitloc.registers[0][0]
        creg_idx = circuit.cregs.index(creg)
        clbit_idx = bitloc.registers[0][1]
        return cls(creg.name, len(creg), creg_idx, clbit_idx)


def serialize_instructions(  # noqa: PLR0912, PLR0915
    circuit: QiskitQuantumCircuit, qubit_index_to_name: dict[int, str], allowed_nonnative_gates: Collection[str] = ()
) -> list[Instruction]:
    """Serialize a quantum circuit into the IQM data transfer format.

    This is IQM's internal helper for :meth:`.IQMBackend.serialize_circuit` that gives slightly more control.
    See :meth:`.IQMBackend.serialize_circuit` for details.

    Args:
        circuit: quantum circuit to serialize
        qubit_index_to_name: Mapping from qubit indices to the corresponding qubit names.
        allowed_nonnative_gates: Names of gates that are converted as-is without validation.
            By default, any gate that can't be converted will raise an error.
            If such gates are present in the circuit, the caller must edit the result to be valid and executable.
            Notably, since IQM transfer format requires named parameters and qiskit parameters don't have names, the
            `i` th parameter of an unrecognized instruction is given the name ``"p<i>"``.

    Returns:
        list of instructions representing the circuit

    Raises:
        ValueError: circuit contains an unsupported instruction or is not transpiled in general

    """
    instructions: list[Instruction] = []
    # maps clbits to the latest "measure" instruction to store its result there
    clbit_to_measure: dict[Clbit, Instruction] = {}
    for circuit_instruction in circuit.data:
        instruction = circuit_instruction.operation
        qubit_names = [qubit_index_to_name[circuit.find_bit(qubit).index] for qubit in circuit_instruction.qubits]
        if instruction.name == "r":
            angle_t = float(instruction.params[0] / (2 * np.pi))
            phase_t = float(instruction.params[1] / (2 * np.pi))
            native_inst = Instruction(name="prx", qubits=qubit_names, args={"angle_t": angle_t, "phase_t": phase_t})
        elif instruction.name == "x":
            native_inst = Instruction(name="prx", qubits=qubit_names, args={"angle_t": 0.5, "phase_t": 0.0})
        elif instruction.name == "rx":
            angle_t = float(instruction.params[0] / (2 * np.pi))
            native_inst = Instruction(name="prx", qubits=qubit_names, args={"angle_t": angle_t, "phase_t": 0.0})
        elif instruction.name == "y":
            native_inst = Instruction(name="prx", qubits=qubit_names, args={"angle_t": 0.5, "phase_t": 0.25})
        elif instruction.name == "ry":
            angle_t = float(instruction.params[0] / (2 * np.pi))
            native_inst = Instruction(name="prx", qubits=qubit_names, args={"angle_t": angle_t, "phase_t": 0.25})
        elif instruction.name == "cz":
            native_inst = Instruction(name="cz", qubits=qubit_names, args={})
        elif instruction.name == "move":
            native_inst = Instruction(name="move", qubits=qubit_names, args={})
        elif instruction.name == "barrier":
            native_inst = Instruction(name="barrier", qubits=qubit_names, args={})
        elif instruction.name == "delay":
            duration = float(instruction.params[0])
            # convert duration to seconds
            unit = instruction.unit
            if unit == "dt":
                duration *= 1e-9  # we arbitrarily pick dt == 1 ns
            elif unit == "s":
                pass
            elif unit == "ms":
                duration *= 1e-3
            elif unit == "us":
                duration *= 1e-6
            elif unit == "ns":
                duration *= 1e-9
            elif unit == "ps":
                duration *= 1e-12
            else:
                raise ValueError(f"Delay: Unsupported unit '{unit}'")
            native_inst = Instruction(name="delay", qubits=qubit_names, args={"duration": duration})
        elif instruction.name == "measure":
            if len(circuit_instruction.clbits) != 1:
                raise ValueError(
                    f"Unexpected: measurement instruction {circuit_instruction} uses multiple classical bits."
                )
            clbit = circuit_instruction.clbits[0]  # always a single-qubit measurement
            mk = str(MeasurementKey.from_clbit(clbit, circuit))
            native_inst = Instruction(name="measure", qubits=qubit_names, args={"key": mk})
            clbit_to_measure[clbit] = native_inst
        elif instruction.name == "reset":
            native_inst = Instruction(name="reset", qubits=qubit_names, args={})
        elif instruction.name == "id":
            continue
        elif instruction.name in allowed_nonnative_gates:
            args = {f"p{i}": param for i, param in enumerate(instruction.params)}
            native_inst = Instruction.model_construct(name=instruction.name, qubits=tuple(qubit_names), args=args)
        else:
            raise ValueError(
                f"Instruction '{instruction.name}' in the circuit '{circuit.name}' is not natively supported. "
                f"You need to transpile the circuit before execution."
            )

        # classically controlled gates (using the c_if method)
        # TODO we do not check anywhere if cc_prx is available for this locus!
        condition = instruction.condition
        if condition is not None:
            if native_inst.name != "prx":
                raise ValueError(
                    f"This backend only supports conditionals on r, x, y, rx and ry gates, not on {instruction.name}"
                )
            native_inst.name = "cc_prx"
            creg, value = condition
            if isinstance(creg, ClassicalRegister):
                if len(creg) != 1:
                    raise ValueError(f"{instruction} is conditioned on multiple bits, this is not supported.")
                if value != 1:
                    raise ValueError(
                        f"{instruction} is conditioned on integer value {value}, only value 1 is supported."
                    )
                clbit = creg[0]
            else:
                clbit = creg  # it is a Clbit

            # Set up feedback routing.
            # The latest "measure" instruction to write to that classical bit is modified, it is
            # given an explicit feedback_key equal to its measurement key.
            # The same feedback_key is given to the controlled instruction, along with the feedback qubit.
            measure_inst = clbit_to_measure[clbit]
            feedback_key = measure_inst.args["key"]
            measure_inst.args["feedback_key"] = feedback_key  # this measure is used to provide feedback
            physical_qubit_name = measure_inst.qubits[0]  # single-qubit measurement
            native_inst.args["feedback_key"] = feedback_key
            native_inst.args["feedback_qubit"] = physical_qubit_name

        instructions.append(native_inst)
    return instructions


def deserialize_instructions(
    instructions: list[Instruction], qubit_name_to_index: dict[str, int], layout: Layout
) -> QiskitQuantumCircuit:
    """Helper function to turn a list of IQM Instructions into a Qiskit QuantumCircuit.

    Args:
        instructions: The gates in the circuit.
        qubit_name_to_index: Mapping from qubit names to their indices, as specified in a backend.
        layout: Qiskit representation of a layout.

    Raises:
        ValueError: Thrown when a given instruction is not supported.

    Returns:
        Qiskit circuit represented by the given instructions.

    """
    # maps measurement key to the corresponding clbit
    mk_to_clbit: dict[str, Clbit] = {}
    # maps feedback key to the corresponding clbit
    fk_to_clbit: dict[str, Clbit] = {}

    # maps creg index to creg in the circuit
    cl_regs: dict[int, ClassicalRegister] = {}

    def register_key(key: str, mapping: dict[str, Clbit]) -> None:
        """Update the classical registers and the given key-to-clbit mapping with the given key."""
        mk = MeasurementKey.from_string(key)
        # find/create the corresponding creg
        creg = cl_regs.setdefault(mk.creg_idx, ClassicalRegister(size=mk.creg_len, name=mk.creg_name))
        # add the key to the given mapping
        if mk.clbit_idx < len(creg):
            mapping[str(mk)] = creg[mk.clbit_idx]
        else:
            raise IndexError(f"{mk}: Clbit index {mk.clbit_idx} is out of range for {creg}.")

    for instr in instructions:
        if instr.name == "measure":
            register_key(instr.args["key"], mk_to_clbit)
            if (key := instr.args.get("feedback_key")) is not None:
                register_key(key, fk_to_clbit)

    # Add resonators
    n_qubits = len(layout.get_physical_bits())
    n_resonators = len(qubit_name_to_index) - n_qubits
    if n_resonators > 0:
        new_qreg = QuantumRegister(n_resonators, "resonators")
        layout.add_register(new_qreg)
        for idx in range(n_resonators):
            layout.add(new_qreg[idx], idx + n_qubits)
    index_to_qiskit_qubit = layout.get_physical_bits()
    # Add an empty Classical register when the original circuit had unused classical registers
    circuit = QiskitQuantumCircuit(
        *layout.get_registers(),
        *(cl_regs.get(i, ClassicalRegister(0)) for i in range(max(cl_regs) + 1 if cl_regs else 0)),
    )
    for instr in instructions:
        locus = [index_to_qiskit_qubit[qubit_name_to_index[q]] for q in instr.qubits]
        if instr.name == "prx":
            angle_t = instr.args["angle_t"] * 2 * np.pi
            phase_t = instr.args["phase_t"] * 2 * np.pi
            circuit.r(angle_t, phase_t, locus[0])
        elif instr.name == "cz":
            circuit.cz(*locus)
        elif instr.name == "move":
            circuit.append(MoveGate(), locus)
        elif instr.name == "measure":
            mk = MeasurementKey.from_string(instr.args["key"])
            circuit.measure(locus[0], mk_to_clbit[str(mk)])
        elif instr.name == "barrier":
            circuit.barrier(*locus)
        elif instr.name == "delay":
            duration = instr.args["duration"]
            circuit.delay(duration, locus, unit="s")  # native delay instructions always use seconds
        elif instr.name == "cc_prx":
            angle_t = instr.args["angle_t"] * 2 * np.pi
            phase_t = instr.args["phase_t"] * 2 * np.pi
            feedback_key = instr.args["feedback_key"]
            # NOTE: 'feedback_qubit' is not needed, because in Qiskit you only have single-qubit measurements.
            circuit.r(angle_t, phase_t, locus[0]).c_if(fk_to_clbit[feedback_key], 1)
        elif instr.name == "reset":
            for qubit in locus:
                circuit.reset(qubit)
        else:
            raise ValueError(f"Unsupported instruction {instr.name} in the circuit.")
    return circuit



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Demonstrates executing a quantum circuit on an IQM quantum computer.

Set the IQM_SERVER_URL environment variable before running this script.
Also, if the server you are running against requires authentication you will also have to set
IQM_AUTH_SERVER, and either IQM_TOKENS_FILE or both of IQM_AUTH_USERNAME and IQM_AUTH_PASSWORD.

E.g.

    export IQM_SERVER_URL="https://example.com/cocos"
    export IQM_AUTH_SERVER="https://example.com/auth"
    export IQM_TOKENS_FILE="/path/to/my/tokens.json"
"""

import os

import cirq
from iqm.cirq_iqm.iqm_sampler import IQMSampler
from iqm.cirq_iqm.optimizers import simplify_circuit
import numpy as np


def fold_func(x: np.ndarray) -> str:
    """Fold the measured bit arrays into strings."""
    return "".join(map(lambda x: chr(x + ord("0")), x))


def demo_run_circuit() -> None:
    """Run a quantum circuit on an IQM quantum computer."""
    a = cirq.NamedQubit("Alice")
    b = cirq.NamedQubit("Bob")
    circuit = cirq.Circuit(cirq.H(a), cirq.CX(a, b), cirq.measure(a, b, key="m"))
    print("Original circuit:\n")
    print(circuit)

    sampler = IQMSampler(os.environ["IQM_SERVER_URL"])

    circuit_routed, _, _ = sampler.device.route_circuit(circuit)
    circuit_decomposed = sampler.device.decompose_circuit(circuit_routed)
    circuit = simplify_circuit(circuit_decomposed)
    print("\nTranspiled and routed circuit:\n")
    print(circuit)
    print("\n")

    # This will send the circuit to the server to be run, and return a cirq.study.Result
    # containing the measurement results.
    results = sampler.run(circuit, repetitions=1000)

    # Sampler results can be accessed several ways
    # For instance, to see the histogram of results
    print(results.histogram(key="m", fold_func=fold_func))

    # Or the data itself
    print(results.data)


if __name__ == "__main__":
    demo_run_circuit()
